/*
 * $Id: Category.cc,v 1.20 2003/04/16 14:11:47 adeniz Exp $
 * This file is part of qualres
 *
 * Copyright (C) 2002  Baker Research, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */
 
/*
 * See the AUTHORS file for a list of people on the gedit Team.  See the
 * ChangeLog files for a list of changes.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "Category.h"
#include "Document.h"
#include "Set.h"
#include <strstream>
#include <iomanip>

/** Structure to hold the context for changing the Category properties.
*/
struct PropertyContext
{
  GtkWidget *parent; ///< Parent of the color selection widget
  GtkWidget *view; ///< View widget.
  GtkTextTag *tag; ///< Tag to update.
  GdkColor *color; ///< Color to update.
  string property; ///< The property of the tag to update.
  string font; ///< Font name to update.
};

/** Structure to hold the context for Coding Coded Text.
*/
struct CodedTextContext
{
  Document *doc; ///< Document associated with the Coded Text.
  GtkTextIter docIterStart; ///< Start of Coded Text in the Document.
  GtkTextIter docIterEnd; ///< End of Coded Text in the Document.
  Category *cat; ///< Text will be Coded into this Category.
  GtkTextBuffer *buffer; ///< Text buffer of Coded Text window.
  GtkWidget *view; ///< View of Coded Text Window.
  bool code; ///< True if selected text is to be Coded, False if Uncoded.
};

// Callbacks.
void categoryCode_cb(GtkWidget *, gpointer);
void categoryUncodeRow_cb(GtkWidget *, gpointer);
void catClearCodedText_cb(GtkWidget *, gpointer);
void cctCodeUncode_cb(GtkWidget *, gpointer);
gboolean categoryUpdate_cb(GtkWidget *, GdkEventCrossing *, gpointer);
gboolean categoryButtonPress_cb(GtkWidget *, GdkEventButton *, gpointer);
void changeColor_cb(GtkWidget *, gpointer);
void changeFont_cb(GtkWidget *, gpointer);

/** Create empty Category.
 */
Category::Category(GtkTextTagTable *t)
{
  init(t);
}

/** Initialize the object

    \param t is the Project tag table

    This method must be called only once for each object.  A tag
    with the name of the object is created and loaded into the
    Project tag table.
*/
void
Category::init(GtkTextTagTable *t)
{ 
  objType = QualResObject::CATEGORY;
  objName = "Category";
  xmlTag = CATEGORY_TAG;
  codedTextStore = NULL;
  projTagTable = t;
  catTag = gtk_text_tag_new(NULL);
  catNameBuffer = gtk_text_buffer_new(projTagTable);
  createWindow(NULL);
  markupInit();
}

/** Initialize the Category markup properties.
*/
void
Category::markupInit()
{ 
  GdkColor color;

  color.red = 0;
  color.green = 0;
  color.blue = 0;
  setFgColor(&color);

  color.red = 0xFFFF;
  color.green = 0xFFFF;
  color.blue = 0xFFFF;
  setBgColor(&color);

  setFont(DEFAULT_FONT);
  underlined = false;
}

/** Destructor.
 */
Category::~Category()
{
  gtk_text_tag_table_remove(projTagTable, catTag);
  if (codedTextStore != NULL)
    g_object_unref(G_OBJECT(codedTextStore));
}

/** Load the Category's font.

      \param f is the new font name

      If the font name is "", this method does nothing.
*/
void
Category::setFont(const string &f)
{

  if (f == "")
    return;

  if (font != f)
    setChanged();

  font = f;

  // Make the changes in the Document.
  g_object_set((gpointer) catTag, "font", font.c_str(), NULL);

  // Make the changes in the Coded Text view.
  if (codedTextRenderer != NULL)
    g_object_set(G_OBJECT(codedTextRenderer),
		 "font", font.c_str(),
		 NULL);

  // Make the changes visible.  This is a work-around.
  // Don't know how to do this; bug in GTK.
}

/** Load the Category's foreground color.

      \param fg is the new foreground color
  */
void
Category::setFgColor(GdkColor *fg)
{
  gboolean success;

  if ((fgColor.red   != fg->red) ||
      (fgColor.blue  != fg->blue) ||
      (fgColor.green != fg->green))
    setChanged();

  fgColor = *fg;

  gdk_colormap_alloc_colors(gdk_colormap_get_system(),
			    &fgColor,
			    1,
			    FALSE,
			    FALSE,
			    &success);

  g_object_set((gpointer) catTag, "foreground-gdk", &fgColor, NULL);
  if (codedTextRenderer != NULL)
    g_object_set(G_OBJECT(codedTextRenderer),
		 "foreground-gdk", &fgColor,
		 NULL);
}

/** Load the Category's background color.

      \param bg is the new background color
  */
void
Category::setBgColor(GdkColor *bg)
{
  gboolean success;

  if ((bgColor.red   != bg->red) ||
      (bgColor.blue  != bg->blue) ||
      (bgColor.green != bg->green))
    setChanged();

  bgColor = *bg;

  gdk_colormap_alloc_colors(gdk_colormap_get_system(),
			    &bgColor,
			    1,
			    FALSE,
			    FALSE,
			    &success);

  g_object_set(G_OBJECT(catTag), "background-gdk", &bgColor, NULL);
  if (codedTextRenderer != NULL)
    g_object_set(G_OBJECT(codedTextRenderer),
		 "background-gdk", &bgColor,
		 NULL);
}

/** Commit the pending new name.

    pendingSetName must have been called.  If the new name is
    different, the object is marked as changed, the window title is
    changed, and the store containing the Category name is updated.
*/
void
Category::commitName()
{
  GtkTextIter start, end;

  Component::commitName();

  // Update the store containing the Category name.
  if (catNameBuffer != NULL)
    {
      gtk_text_buffer_set_text(catNameBuffer,
			       (getPrettyName()).c_str(),
			       -1);
      gtk_text_buffer_get_bounds(catNameBuffer,
				 &start,
				 &end);
      gtk_text_buffer_apply_tag(catNameBuffer,
				catTag,
				&start,
				&end);
    }
}

/** Return number of Documents.

\returns number of Documents
*/
guint
Category::nDocuments() const
{
  map<Document *, DocCodedText>::const_iterator iDoc;
  guint nDocs;

  // Count only uncut Documents.
  for (nDocs =  0, iDoc = docCodedTextMap.begin();
       iDoc != docCodedTextMap.end();
       iDoc++)
    {
      if (!(iDoc->second).cut)
	nDocs++;
    }
  return(nDocs);
}

/** Get the text to print.

    \returns a pointer to allocated memory containing the text to print.
    This memory must be freed by g_free().
*/
gchar *
Category::getTextToPrint() const
{
  gchar *text = "", *docLine, *speakerID, *codedText;
  GtkTreeIter docIter, codedTextIter;
  gboolean validParent, validCodedText;
  gint indx;
  string codedTextStr;

  g_return_val_if_fail((codedTextStore != NULL), NULL);

  validParent = gtk_tree_model_get_iter_first(GTK_TREE_MODEL(codedTextStore),
					      &docIter);

  while (validParent)
    {
      // Get the document data.
      gtk_tree_model_get(GTK_TREE_MODEL(codedTextStore),
			 &docIter,
			 Category::DOCUMENT_LINE, &docLine,
			 -1);
      text = g_strconcat(((text == NULL) ? "" : text),
			 ((docLine == NULL) ? "" : docLine),
			 "\n",
			 NULL);
      g_free(docLine);

      // Get the Coded Text data.
      validCodedText = gtk_tree_model_iter_children
	(GTK_TREE_MODEL(codedTextStore),
	 &codedTextIter,
	 &docIter);
      while (validCodedText)
	{
	  // Get the Coded Text data.
	  gtk_tree_model_get(GTK_TREE_MODEL(codedTextStore),
			     &codedTextIter,
			     Category::DOCUMENT_LINE, &docLine,
			     Category::SPEAKER_ID, &speakerID,
			     Category::CODED_TEXT, &codedText,
			     -1);

	  // Add padding after each occurance of "\n" in the Coded Text.
	  codedTextStr = codedText;
	  indx = 0;
	  while ((indx = codedTextStr.find(NEWLINE_CHAR, indx)) !=
		 string::npos)
	    codedTextStr.insert(++indx, PRINT_LINE_PAD);

	  text = g_strconcat(((text == NULL) ? "" : text),
			     PRINT_NUMBER_PAD,
			     ((docLine == NULL) ? "" : docLine),
			     NEWLINE_STRING,
			     PRINT_ID_PAD,
			     ((speakerID == NULL) ? "" : speakerID),
			     PRINT_ID_LINE_PAD,
			     codedTextStr.c_str(),
			     "\n\n",
			     NULL);
	  g_free(docLine);
	  g_free(speakerID);
	  g_free(codedText);

      // Get the next Coded Text iterator.
	  validCodedText = gtk_tree_model_iter_next
	    (GTK_TREE_MODEL(codedTextStore),
	     &codedTextIter);
	}

      // Get the next Document iterator.
      validParent = gtk_tree_model_iter_next(GTK_TREE_MODEL(codedTextStore),
					     &docIter);
    }

  return(text);
}

/** Get the Component name labeled with "Category:  ".

    \returns left header of print output (free with g_free)
**/
gchar *
Category::getLabeledName() const
{
  return(g_strconcat("Category:  ", (getPrettyName()).c_str(), NULL));
}

/** Code the text in a Document.

    \param document is the Document

    \param start is the start of the text to code

    \param end is the end of the text to code

    \param loadUpdate specifies whether to load or update Coded Text
    container

*/
void
Category::codeDocText(Document *doc,
		      GtkTextIter *start,
		      GtkTextIter *end,
		      Category::LoadUpdate loadUpdate)
{
  GtkTextBuffer *buffer = doc->getTextBuffer();

  DocCodedText *docCodedText;
  CodedText codedText;
  vector<CodedText>::iterator iter; // Iter for Coded Text container

  gboolean oldDocument;

  g_return_if_fail(codedTextStore != NULL);

  // Return if the iterators are equal.
  if (gtk_text_iter_compare(start, end) == 0)
      return;

  // Has this Document been coded before?
  if (docCodedTextMap.count(doc) == 0)
      oldDocument = FALSE;
  else
      oldDocument = TRUE;

  // Find the DocCodedText structure for this Document.
  docCodedText = &(docCodedTextMap[doc]);
  if (oldDocument == FALSE)
    {
      docCodedText->nextID = Category::INIT_ID;
      docCodedText->cut = FALSE;

      // Initialize the Document iterator.
      gtk_tree_store_append(codedTextStore,
			    &(docCodedText->iter),
			    NULL);

      // Load the Document pointer.
      docCodedText->doc = doc;

      // Load the Document name into the model.
      gtk_tree_store_set(codedTextStore,
			 &(docCodedText->iter),
			 Category::DOCUMENT_LINE,
			 (doc->getPrettyName()).c_str(),
			 Category::DOCUMENT_NAME,
			 (doc->getName()).c_str(),
			 -1);
    }

  // Determine the limits of the Coded Text after merging with
  // existing Coded Text segments.  This method also uncodes any Coded
  // Text that is merged, and untags the Document text.
  if (loadUpdate == Category::UPDATE)
    {
      if (!mergeCodedText(docCodedText, start, end))
	return;
    }

  // Initialize the CodedText for this entry.
  codedText.ID = docCodedText->nextID++;
  codedText.docStart = gtk_text_buffer_create_mark(buffer,
						   NULL,
						   start,
						   FALSE);
  codedText.docEnd = gtk_text_buffer_create_mark(buffer,
						 NULL,
						 end,
						 TRUE);

  // Get an iterator for this row in the Category display.
  gtk_tree_store_append(codedTextStore,
			&(codedText.catIter),
			&(docCodedText->iter));

  // Add to docCodedText for this Document.
  iter = docCodedText->codedTextVector.begin();
  docCodedText->codedTextVector.insert(iter, codedText);

  // Apply the tag.
  gtk_text_buffer_apply_tag(buffer, catTag, start, end);

  // Request the Document to add a Category label.
  doc->addCategoryLabel(this, start);

  // Remove any Coded Text marked for deletion.
  if (loadUpdate == Category::UPDATE)
    pruneCodedText(docCodedText);

  // Display the Coded Text in the Document store on the
  // Category buffer.
  displayCodedText(doc,
		   start,
		   end,
		   &(codedText.catIter),
		   codedText.ID);

  // Update the number of Documents Coded in this Category.
    if ((oldDocument == FALSE) && (loadUpdate != Category::LOAD))
    {
      GtkTreeIter catIter;
      map<const string, Set *>::iterator iSet;

      // Update the Project store.
      componentProjectIter(&catIter);
      gtk_tree_store_set(selectCatStore,
			 &catIter,
			 QualResObject::N_DOCS, nDocuments(),
			 -1);

      // Inform all Sets.
      for (iSet = projectSetMap->begin(); iSet != projectSetMap->end(); iSet++)
	{
	  (iSet->second)->propertiesHaveChanged(this);
	}
    }

  setChanged();
}

/** Merge the text to be Coded with existing Coded Text.

    \param docCodedText contains the existing Coded Text.

    \param start is the start of the text to be Coded.

    \param end is the end of the text to be Coded.

    \return true if text is to be Coded, false otherwise.

    This method checks each existing Coded Text for overlap with the
    text to be Coded.  If there is any overlap, that existing Coded
    Text is marked for removal (using the ID field) and the limits of
    the text to be Coded are expanded accordingly.  This assumes that
    all the existing Coded Text are disjoint--no overlaps.

*/
bool
Category::mergeCodedText(DocCodedText *docCodedText,
			 GtkTextIter *start,
			 GtkTextIter *end)
{
  GtkTextIter oldStart, oldEnd; // For existing Coded Text
  GtkTextBuffer *textBuffer;
  vector<CodedText>::iterator iText;

  // Return if the iterators are equal.
  if (gtk_text_iter_compare(start, end) == 0)
    return(false);

  textBuffer = (docCodedText->doc)->getTextBuffer();

  // Check each existing Coded Text.
  for (iText = (docCodedText->codedTextVector).begin();
       iText != (docCodedText->codedTextVector).end();
       iText++)
    {
      // Get limits of existing Coded Text.
      gtk_text_buffer_get_iter_at_mark(textBuffer,
				       &oldStart,
				       iText->docStart);
      gtk_text_buffer_get_iter_at_mark(textBuffer,
				       &oldEnd,
				       iText->docEnd);

      // If there's no overlap, do nothing.
      if ((gtk_text_iter_compare(&oldEnd, start) < 0) ||
	  (gtk_text_iter_compare(end , &oldStart) < 0))
	continue;

      // Start of text to Code is within existing Coded Text.
      if ((gtk_text_iter_compare(&oldStart, start) <= 0) &&
	  (gtk_text_iter_compare(start, &oldEnd) <= 0))
      {
	// If end of text to Code is also within the existing Coded
	// Text, then we should not Code the text.  Also, none of the
	// other Coded Text was changed, because by assumption they
	// are all disjoint.  We're done.
	if (gtk_text_iter_compare(end, &oldEnd) <= 0)
	  return(false);
	else
	  // The end of text to Code is after the end of the existing
	  // Coded Text.
	  {
	    //  Extend the text to Code
	    *start = oldStart;

	    // Flag this Coded Text for deletion.
	    iText->ID = Category::PENDING_UNCODE;
	  }
	continue;
      }
      // Start of text to Code is before the existing Coded Text.
      if (gtk_text_iter_compare(start, &oldStart) <= 0)
      {
	// End of text to Code is after the existing Coded Text.
	if (gtk_text_iter_compare(end, &oldEnd) >= 0)
	  {
	    // The Coded Text is contained entirely within the text to Code.
	    // Flag this Coded Text for deletion.
	    iText->ID = Category::PENDING_UNCODE;
	  }
	else
	  // The end of text to Code is within the Coded Text.
	  {
	  // Extend the text to code to the end of the Coded Text.
	    *end = oldEnd;

	    // Flag this Coded Text for deletion.
	    iText->ID = Category::PENDING_UNCODE;
	  }
      }
    }
  // The text to Code should be Coded.
  return(true);
}

/** Remove Coded Text that is zero-length or marked for deletion.

    \param docCodedText contains the Coded Text to prune.

    Remove means that the CodedText structure is deleted, and the
    corresponding text in the Document text buffer is uncoded.

*/
void
Category::pruneCodedText(DocCodedText *docCodedText)
{
  GtkTextBuffer *textBuffer;
  GtkTextIter start, end;
  vector<CodedText>::iterator iText;

  textBuffer = (docCodedText->doc)->getTextBuffer();

  // Check each existing Coded Text.
  for (iText = (docCodedText->codedTextVector).end();
       iText != (docCodedText->codedTextVector).begin();
       )
    {
      iText--;

      // Get the iterators for comparison.
      gtk_text_buffer_get_iter_at_mark(textBuffer,
				       &start,
				       iText->docStart);
      gtk_text_buffer_get_iter_at_mark(textBuffer,
				       &end,
				       iText->docEnd);

      // If the Coded Text has zero length, it should be uncoded.
      if (gtk_text_iter_equal(&start, &end))
	iText->ID = Category::PENDING_UNCODE;

      // Uncode if the Coded Text is to be uncoded.
      if (iText->ID == Category::PENDING_UNCODE)
	{
	  // Remove the tag.
	  gtk_text_buffer_remove_tag(textBuffer,
				     catTag,
				     &start,
				     &end);

	  // Remove the row.
	  if (codedTextStore != NULL)
	    gtk_tree_store_remove(codedTextStore, &(iText->catIter));

	  // Remove the text marks.
	  gtk_text_buffer_delete_mark(textBuffer,
				      iText->docStart);
	  gtk_text_buffer_delete_mark(textBuffer,
				      iText->docEnd);

	  // Inform the Document.
	  (docCodedText->doc)->removeCategoryLabel(this, &start);

	  // Remove the CodedText.
	  docCodedText->codedTextVector.erase(iText);

	  setChanged();
	}
    }
}

/** Inform the object that a component is about to be deleted.

    \param comp is the component

    If comp is a Document, this method removes any record of the Coded
    Text.
*/
void
Category::pendingDelete(QualResObject *comp)
{
  Document *doc = (Document *) comp;

  // Update selected Category.
  if ((Category *) comp == selectedCat)
    setSelectedCategory("");

  // If this Category is to be Deleted, hide the Selection Windows.
  if ((Category *) comp == this)
    {
      hideAllSelectWindows();
      return;
    }

  // Is this a Document that has Coded Text in this Category?
  if (docCodedTextMap.count(doc) != 0)
    {
      DocCodedText *docCodedText = &(docCodedTextMap[doc]);
      vector<CodedText>::iterator iText;

      // Flag each row to be deleted.
      for (iText = (docCodedText->codedTextVector).begin();
	   iText != (docCodedText->codedTextVector).end();
	   iText++)
	iText->ID = Category::PENDING_UNCODE;

      // Remove rows to be deleted and update the Category view.
      loadCodedTextStore(Category::UPDATE);
    }
}

/** Inform the object that a component is about to be cut.

    \param comp is the component
*/
void
Category::pendingCut(QualResObject *comp)
{
  Document *doc;

  // Update selected Category.
  if ((DocCat *) comp == selectedCat)
    setSelectedCategory("");

  // If this is the component to be cut, clear the properties of the
  // tag, and hide any selection windows.
  if ((Category *) comp == this)
    {
      gchar *fontName;
      GdkColor color;
      GtkTextTag *tempTag;

      // Create a new tag, and copy its properties to catTag.
      tempTag = gtk_text_tag_new(NULL);

      // Set the font to default.
      g_object_get((gpointer) tempTag,
		   "font",
		   &fontName,
		   NULL);
      g_object_set((gpointer) catTag,
		   "font",
		   fontName,
		   NULL);
      g_free(fontName);

      // Clear the Category name buffer.
      if (catNameBuffer != NULL)
	gtk_text_buffer_set_text(catNameBuffer, "", -1);

      // Set the colors to the default.
      color.red = 0xFFFF;
      color.green = 0xFFFF;
      color.blue = 0xFFFF;
      g_object_set(G_OBJECT(catTag), "background-gdk", &color, NULL);

      color.red = 0;
      color.green = 0;
      color.blue = 0;
      g_object_set(G_OBJECT(catTag), "foreground-gdk", &color, NULL);

      g_object_unref(G_OBJECT(tempTag));

      hideAllSelectWindows();
      return;
    }

  // If a Document is about to be cut...
  doc = (Document *) comp;
  if (docCodedTextMap.count(doc) != 0)
    {
      DocCodedText *docCodedText;
      map<Document *, DocCodedText>::iterator iDoc;
      vector<CodedText>::iterator iText;

      // Delete any previously cut DocCodedText element.
      for (iDoc = docCodedTextMap.begin();
	   iDoc != docCodedTextMap.end();
	   iDoc++)
	{
	  if ((iDoc->second).cut)
	      docCodedTextMap.erase(iDoc);
	}

      // Flag this DocCodedText element as Cut.
      docCodedText = &(docCodedTextMap[doc]);
      docCodedText->cut = TRUE;

      if (codedTextStore != NULL)
	{
	  // Delete each row in the Category view.
	  for (iText = (docCodedText->codedTextVector).begin();
	       iText != (docCodedText->codedTextVector).end();
	       iText++)
	    gtk_tree_store_remove(codedTextStore, &(iText->catIter));

      // Delete the Document row in the Category view.
	  gtk_tree_store_remove(codedTextStore, &(docCodedText->iter));
	}

      // Update the number of Documents Coded in this Category.
      GtkTreeIter catIter;
      map<const string, Set *>::iterator iSet;

      // Update the Project Store.
      componentProjectIter(&catIter);
      gtk_tree_store_set(selectCatStore,
			 &catIter,
			 QualResObject::N_DOCS, nDocuments(),
			 -1);

      // Inform all Sets.
      for (iSet = projectSetMap->begin(); iSet != projectSetMap->end(); iSet++)
	{
	  (iSet->second)->propertiesHaveChanged(this);
	}
    }
}

/** Inform the object that a component is about to be Pasted.

    \param comp is the component
*/
void
Category::pendingPaste(QualResObject *comp)
{
  Document *doc;

  g_return_if_fail(codedTextStore != NULL);

  // If this is the component to be Pasted, restore the properties of the
  // tag.
  if ((Category *) comp == this)
    {
      setFgColor(&fgColor);
      setBgColor(&bgColor);
      setFont(font);
    }  
  // If a Document that has Coded Text in this Category is about to be
  // pasted...
  doc = (Document *) comp;
  if (docCodedTextMap.count(doc) != 0)
    {
      GtkTextIter startIter, endIter;
      DocCodedText *docCodedText = &(docCodedTextMap[doc]);
      vector<CodedText>::iterator iText;

      // Flag this DocCodedText element as Uncut.
      docCodedText->cut = FALSE;

      // Initialize the Document iterator.
      gtk_tree_store_append(codedTextStore,
			    &(docCodedText->iter),
			    NULL);

      // Load the Document pointer.
      docCodedText->doc = doc;

      // Load the Document name into the model.
      gtk_tree_store_set(codedTextStore,
			 &(docCodedText->iter),
			 Category::DOCUMENT_LINE,
			 (doc->getPrettyName()).c_str(),
			 Category::DOCUMENT_NAME,
			 (doc->getName()).c_str(),
			 -1);

      // Update the number of Documents column in the Project tree view.
      GtkTreeIter catIter;

      componentProjectIter(&catIter);

      gtk_tree_store_set(selectCatStore,
			 &catIter,
			 QualResObject::N_DOCS, nDocuments(),
			 -1);

      // Create a row for each Coded Text.
      for (iText = (docCodedText->codedTextVector).begin();
	   iText != (docCodedText->codedTextVector).end();
	   iText++)
	{
	  // Get an iterator for this row in the Category display.
	  gtk_tree_store_append(codedTextStore,
				&(iText->catIter),
				&(docCodedText->iter));

	  // Get the start and end text buffer iterators.
	  gtk_text_buffer_get_iter_at_mark(doc->getTextBuffer(),
					   &startIter,
					   iText->docStart);
	  gtk_text_buffer_get_iter_at_mark(doc->getTextBuffer(),
					   &endIter,
					   iText->docEnd);

	  // Display the Coded Text in the Document store on the
	  // Category buffer.
	  displayCodedText(doc,
			   &startIter,
			   &endIter,
			   &(iText->catIter),
			   iText->ID);
	}
    }
}

/** Remove all the Coded Text from the Category.
*/
void
Category::clearCodedText()
{
  map<Document *, DocCodedText>::iterator iDocCodedText;
  DocCodedText *docCodedText;
  vector<CodedText>::iterator iCodedText;

  // For each Document...
  for (iDocCodedText = docCodedTextMap.begin();
       iDocCodedText != docCodedTextMap.end();
       iDocCodedText++)
    {
      docCodedText = &(iDocCodedText->second);
      // Mark each Coded Text for deletion.
      for (iCodedText = (docCodedText->codedTextVector).begin();
	   iCodedText != (docCodedText->codedTextVector).end();
	   iCodedText++)
	iCodedText->ID = Category::PENDING_UNCODE;

      // Remove the Coded Text for this Document.
      pruneCodedText(docCodedText);
    }
}

/** Uncode the row selected in a Category window.
*/
void
Category::uncodeCatSelection()
{
  GtkTreeModel *model;
  GtkTreeSelection *selection;
  GtkTreeIter uncodeIter, docIter;
  DocCodedText *docCodedText;
  Document *doc;
  GtkTextIter start, end;
  gchar *name;
  guint ID;

  vector<CodedText>::iterator iText;

  g_return_if_fail(codedTextStore != NULL);

  // Get the selected row.
  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(codedTextView));

  // Get the iterator for the selected row.
  if (!gtk_tree_selection_get_selected(selection,
				       &model,
				       &uncodeIter))
    //  There is no row selected.
    return;

  // Find the Document.
  if (!gtk_tree_model_iter_parent(GTK_TREE_MODEL(codedTextStore),
				  &docIter,
				  &uncodeIter))
    // There is no Document.
    return;

  gtk_tree_model_get(GTK_TREE_MODEL(codedTextStore),
		     &docIter,
		     Category::DOCUMENT_NAME, &name,
		     -1);

  doc = (Document *) (*projectDocMap)[name];
  g_free(name);
  gtk_tree_model_get(GTK_TREE_MODEL(codedTextStore),
		     &uncodeIter,
		     Category::CODED_TEXT_ID, &ID,
		     -1);

  // Find and remove the CodedText element that matches this element.
  docCodedText = &(docCodedTextMap[doc]);
  for (iText = docCodedText->codedTextVector.begin();
       iText != docCodedText->codedTextVector.end();
       iText++)
    {
      if (iText->ID == ID)
	{
	  // Get the iterators in the Document buffer.
	  gtk_text_buffer_get_iter_at_mark(doc->getTextBuffer(),
					   &start,
					   iText->docStart);
	  gtk_text_buffer_get_iter_at_mark(doc->getTextBuffer(),
					   &end,
					   iText->docEnd);

	  // Uncode the text in the Document.
	  gtk_text_buffer_remove_tag(doc->getTextBuffer(),
				     catTag,
				     &start,
				     &end);

	  // Inform the Document.
	  (docCodedText->doc)->removeCategoryLabel(this, &start);

	  // Delete this element.
	  docCodedText->codedTextVector.erase(iText);
	  break;
	}
    }

  // Remove the row.
  gtk_tree_store_remove(codedTextStore, &uncodeIter);

  // If there is no Coded Text, remove the Document line and erase
  // this docCodedTextMap element.
  if ((docCodedText->codedTextVector).size() == 0)
    {
      GtkTreeIter catIter;

      gtk_tree_store_remove(codedTextStore, &(docCodedText->iter));
      docCodedTextMap.erase(doc);

      // Update the number of Documents column.
      componentProjectIter(&catIter);
      map<const string, Set *>::iterator iSet;

      gtk_tree_store_set(selectCatStore,
			 &catIter,
			 QualResObject::N_DOCS, nDocuments(),
			 -1);

      // Inform all Sets.
      for (iSet = projectSetMap->begin(); iSet != projectSetMap->end(); iSet++)
	{
	  (iSet->second)->propertiesHaveChanged(this);
	}
    }

  setChanged();
}

/** Code the text selected in a Category.

    \param name is the name of the Category

    \return true or false, indicating success or failure
*/
bool
Category::codeCatSelection(string &catName)
{
  return(true);
}

/** Remove redundant coded selections

    \return true or false, indicating success or failure
*/
bool
Category::removeRedundant()
{
  return(true);
}

/** Write to an XML element.

    \param xmlDoc is the XML document

    \return a pointer to the XML element

    The objects creates an XML element, and writes its persistent data
    to it.  For example, the buffer is saved, but the cursor positon
    is not.
*/
GdomeElement *
Category::writeXML(GdomeDocument *xmlDoc)
  throw(FileException)
{
  GdomeElement *elem, *docCodedTextElem, *codedTextElem;
  GdomeText *descNode;
  GdomeNode *result;
  GdomeException exc;
  GdomeDOMString *tempStr, *tempVal;

  // Coded Text structure pointers.
  DocCodedText *docCodedText;

  Document *doc;
  GtkTextIter startIter, endIter;
  gint startOffset, endOffset;

  // Iterators.
  map<Document *, DocCodedText>::iterator iDocCodedText;
  vector<CodedText>::iterator iCodedText;

  ostrstream colorAttr;
  ostrstream tempStream;

  // Create element.
  tempStr = gdome_str_mkref(xmlTag);
  elem = gdome_doc_createElement(xmlDoc, tempStr, &exc);
  gdome_str_unref(tempStr);
  if (elem == NULL)
    throw FileException();

  try
    {
      // Set the attributes.
      setAttribute(elem, FONT_ATTR, (gchar *) font.c_str());

      // Save the color attributes.
      colorAttr << fgColor.red << " " << fgColor.green << " " << fgColor.blue
		<< ends;
      setAttribute(elem, FG_RGB_ATTR, (gchar *) colorAttr.str());
      colorAttr.rdbuf()->freeze(0);
      colorAttr.seekp(0, ios::beg);

      colorAttr << bgColor.red << " " << bgColor.green << " " << bgColor.blue
		<< ends;
      setAttribute(elem, BG_RGB_ATTR, (gchar *) colorAttr.str());
      colorAttr.rdbuf()->freeze(0);

      // Create and write the name element.
      setTextInElement(xmlDoc, elem,
		       NAME_TAG,
		       (gchar *) (getName()).c_str());

      // Create and write the description element.
      setTextInElement(xmlDoc, elem,
		       DESCRIPTION_TAG,
		       (gchar *) (getDesc()).c_str());
    }
  catch(FileException)
    {
      gdome_el_unref(elem, &exc);
      throw FileException();
    }

  // Save all Coded Text for each Document.
  for (iDocCodedText = docCodedTextMap.begin();
       iDocCodedText != docCodedTextMap.end();
       iDocCodedText++)
    {
      docCodedText = &(iDocCodedText->second);
      doc = docCodedText->doc;

      // Skip this Document if it has been Cut.
      if (docCodedText->cut)
	continue;

      // Remove zero-length Coded Text.
      pruneCodedText(docCodedText);

      // Create the XML element for the Document Coded Text.
      tempStr = gdome_str_mkref(DOCUMENT_CODED_TEXT_TAG);
      docCodedTextElem = gdome_doc_createElement(xmlDoc, tempStr, &exc);
      gdome_str_unref(tempStr);
      if (docCodedTextElem == NULL)
	{
	  gdome_el_unref(elem, &exc);
	  throw FileException();
	}

      try
	{
	  // Set the Document name.
	  setTextInElement(xmlDoc,
			   docCodedTextElem,
			   NAME_TAG,
			   (gchar *) (doc->getName()).c_str());
	}
      catch(FileException)
	{
	  gdome_el_unref(docCodedTextElem, &exc);
	  gdome_el_unref(elem, &exc);
	  throw FileException();
 	}

      // Save all Coded Text for this Document.
      for (iCodedText = (docCodedText->codedTextVector).begin();
	   iCodedText != (docCodedText->codedTextVector).end();
	   iCodedText++)
	{
	  // Get the character offsets.
	  gtk_text_buffer_get_iter_at_mark(doc->getTextBuffer(),
					   &startIter,
					   iCodedText->docStart);
	  startOffset = gtk_text_iter_get_offset(&startIter);
	  gtk_text_buffer_get_iter_at_mark(doc->getTextBuffer(),
					   &endIter,
					   iCodedText->docEnd);
	  endOffset = gtk_text_iter_get_offset(&endIter);

	  // Create the XML element.
	  tempStr = gdome_str_mkref(CODED_TEXT_TAG);
	  codedTextElem = gdome_doc_createElement(xmlDoc, tempStr, &exc);
	  gdome_str_unref(tempStr);
	  if (codedTextElem == NULL)
	    {
	      gdome_el_unref(docCodedTextElem, &exc);
	      gdome_el_unref(elem, &exc);
	      throw FileException();
	    }

	  // Load the start and end offset attributes.
	  tempStream << startOffset << " " << endOffset << ends;
	  try
	    {
	      setAttribute(codedTextElem,
			   CODED_TEXT_LIMITS_ATTR,
			   (gchar *) tempStream.str());
	    }
	  catch(FileException)
	    {
	      tempStream.rdbuf()->freeze(0);
	      tempStream.seekp(0, ios::beg);
	      gdome_el_unref(codedTextElem, &exc);
	      gdome_el_unref(docCodedTextElem, &exc);
	      gdome_el_unref(elem, &exc);
	      throw FileException();
	    }
	  tempStream.rdbuf()->freeze(0);
	  tempStream.seekp(0, ios::beg);

	  // Append this element to the childs list of docCodedTextElem.
	  result = gdome_el_appendChild(docCodedTextElem,
					(GdomeNode *) codedTextElem,
					&exc);
	  if (result != (GdomeNode *) codedTextElem)
	    {
	      gdome_el_unref(codedTextElem, &exc);
	      gdome_el_unref(docCodedTextElem, &exc);
	      gdome_el_unref(elem, &exc);
	      throw FileException();
	    }
	  gdome_el_unref(codedTextElem, &exc);
	}

      // Append the docCodedTextElem to the childs list of the
      // Category element.
      result = gdome_el_appendChild(elem,
				    (GdomeNode *) docCodedTextElem,
				    &exc);
      if (result != (GdomeNode *)docCodedTextElem)
	{
	      gdome_el_unref(docCodedTextElem, &exc);
	      gdome_el_unref(elem, &exc);
	      throw FileException();
	}

      gdome_el_unref(docCodedTextElem, &exc);
    }

  setSaved();
  return(elem);
}

/** Read from an XML element.

    \param elem is the XML element containing the object data
*/
bool
Category::readXML(GdomeElement *elem)
  throw(FileException)
{
  GdomeNodeList *docCodedTextElemList, *codedTextElemList;
  gint nDocCodedTextElems; // Number of children of this element
  gint nCodedTextElems; // Number of children of each docCodedText element
  GdomeException exc;
  GdomeElement *docCodedTextElem, *codedTextElem; // XML elements
  GdomeDOMString *tempStr;
  string tempString;
  Document *doc;
  gint startOffset, endOffset; // Start and end offsets of the Coded Text
  GtkTextIter startIter, endIter; // Start and end iterators
  GdkColor color;
  gint i, j; // Index

  // Get and load the attributes.
  setFont(getAttribute(elem, FONT_ATTR));

  tempString = getAttribute(elem, FG_RGB_ATTR);
  istrstream tempFg(tempString.c_str());
  tempFg >> color.red >> color.green >> color.blue;
  setFgColor(&color);
  
  tempString = getAttribute(elem, BG_RGB_ATTR);
  istrstream tempBg(tempString.c_str());
  tempBg >> color.red >> color.green >> color.blue;
  setBgColor(&color);
  
  // Get and load the name.
  pendingSetName(getTextFromElement(elem, NAME_TAG));
  commitName();

  // Get and load the description.
  setDesc(getTextFromElement(elem, DESCRIPTION_TAG));

  // Get child elements of the Category.
  docCodedTextElemList = gdome_el_childNodes(elem, &exc);
  if (docCodedTextElemList == NULL)
    throw FileException();
  nDocCodedTextElems = gdome_nl_length(docCodedTextElemList, &exc);

  for (i = 0; i < nDocCodedTextElems; i++)
    {
      // Get element.
      docCodedTextElem = (GdomeElement *)
	gdome_nl_item(docCodedTextElemList, i, &exc);
      if (docCodedTextElem == NULL)
	{
	  gdome_nl_unref(docCodedTextElemList, &exc);
	  throw FileException();
	}

      // If this is not the node with the request tag, go on to the
      // next child.
      tempStr = gdome_el_nodeName(docCodedTextElem, &exc);
      if (exc)
	{
	  gdome_nl_unref(docCodedTextElemList, &exc);
	  throw FileException();
	}
      tempString = tempStr->str;
      gdome_str_unref(tempStr);

      if (tempString != DOCUMENT_CODED_TEXT_TAG)
	{
	  gdome_el_unref(docCodedTextElem, &exc);
	  continue;
	}

      // Get the Document.
      try
	{
	  doc = (*projectDocMap)[getTextFromElement(docCodedTextElem,
						    NAME_TAG)];
	}
      catch(FileException)
	{
	  gdome_el_unref(docCodedTextElem, &exc);
	  gdome_nl_unref(docCodedTextElemList, &exc);
	  throw FileException();
	}

      // Get the list of Coded Text elements.
      codedTextElemList = gdome_el_childNodes(docCodedTextElem, &exc);
      if (codedTextElemList == NULL)
	{
	  gdome_el_unref(docCodedTextElem, &exc);
	  gdome_nl_unref(docCodedTextElemList, &exc);
	  throw FileException();
	}
      nCodedTextElems = gdome_nl_length(codedTextElemList, &exc);

      // Load each Coded Text.
      for (j = 0; j < nCodedTextElems; j++)
	{
	  // Get element.
	  codedTextElem = (GdomeElement *)
	    gdome_nl_item(codedTextElemList, j, &exc);
	  if (codedTextElem == NULL)
	    {
	      gdome_nl_unref(codedTextElemList, &exc);
	      gdome_el_unref(docCodedTextElem, &exc);
	      gdome_nl_unref(docCodedTextElemList, &exc);
	      throw FileException();
	    }

	  // If this is not the node with the request tag, go on to the
	  // next child.
	  tempStr = gdome_el_nodeName(codedTextElem, &exc);
	  if (exc)
	    {
	      gdome_el_unref(codedTextElem, &exc);
	      gdome_nl_unref(codedTextElemList, &exc);
	      gdome_el_unref(docCodedTextElem, &exc);
	      gdome_nl_unref(docCodedTextElemList, &exc);
	      throw FileException();
	    }
	  tempString = tempStr->str;
	  gdome_str_unref(tempStr);

	  if (tempString != CODED_TEXT_TAG)
	    {
	      gdome_el_unref(codedTextElem, &exc);
	      continue;
	    }

	  // Get the start and end offsets.
	  try
	    {
	      tempString = getAttribute(codedTextElem, CODED_TEXT_LIMITS_ATTR);
	    }
	  catch(FileException)
	    {
	      gdome_el_unref(codedTextElem, &exc);
	      gdome_nl_unref(codedTextElemList, &exc);
	      gdome_el_unref(docCodedTextElem, &exc);
	      gdome_nl_unref(docCodedTextElemList, &exc);
	      throw FileException();
	    }	    

	  istrstream tempStream(tempString.c_str());
	  tempStream >> startOffset >> endOffset;

	  // Get the start and end iterators.
	  gtk_text_buffer_get_iter_at_offset(doc->getTextBuffer(),
					     &startIter,
					     startOffset);
	  gtk_text_buffer_get_iter_at_offset(doc->getTextBuffer(),
					     &endIter,
					     endOffset);
	  // Code the text.
	  codeDocText(doc, &startIter, &endIter, Category::LOAD);

	  gdome_el_unref(codedTextElem, &exc);
	}

      gdome_nl_unref(codedTextElemList, &exc);
      gdome_el_unref(docCodedTextElem, &exc);
    }

  gdome_nl_unref(docCodedTextElemList, &exc);

  setSaved();
  return(TRUE);
}

/** Search for and code text into the selected Category.

    \param search is the string to search for
    \param searchType is the type of search
    \param cat is the selected DocCat
*/
void
Category::codeSearchText(string &search,
			 DocCat::SearchStore searchType,
			 Category *destCat)
{

  DocCodedText *docCodedText;
  Document *doc;

  // Start and end iterators for Coded Text.
  GtkTextIter startIter, endIter;

  // Index.
  map<Document *, DocCodedText>::iterator iDoc;
  vector<CodedText>::iterator iText;

  // For each Document that has text coded in this Category...
  for (iDoc = docCodedTextMap.begin(); iDoc != docCodedTextMap.end(); iDoc++)
    {
      docCodedText = &(iDoc->second);
      doc = docCodedText->doc;

      // Skip this Document if it has been Cut.
      if (docCodedText->cut)
	continue;

      // Search each Coded Text.
      for (iText = (docCodedText->codedTextVector).begin();
	   iText != (docCodedText->codedTextVector).end();
	   iText++)
	{
	  // Obtain the iterators of the Coded Text.
	  gtk_text_buffer_get_iter_at_mark(doc->getTextBuffer(),
					   &startIter,
					   iText->docStart);
	  gtk_text_buffer_get_iter_at_mark(doc->getTextBuffer(),
					   &endIter,
					   iText->docEnd);

	  // Search and code.
	  codeSearchCodedText(doc,
			      search,
			      searchType,
			      &startIter,
			      &endIter,
			      destCat);
	}
    }

}

/** Search for and code Coded Text into the selected Category.

    \param doc is the Document the Coded Text belongs to
    \param search is the string to search for
    \param searchType is the type of search
    \param startIter points to the beginning of the text to search
    \param endIter points to the end of the text to search
    \param cat is the selected DocCat
*/
void
Category::codeSearchCodedText(Document *doc,
			      string &search,
			      DocCat::SearchStore searchType,
			      GtkTextIter *startIter,
			      GtkTextIter *endIter,
			      Category *destCat)
{
  gint startSearch, searchLength = search.length();
  string textToSearch;

  GtkTextIter matchStart, matchEnd; // Boundaries of found text

  textToSearch = gtk_text_buffer_get_text(doc->getTextBuffer(),
					  startIter, endIter,
					  TRUE);

  // Code while a match is found.
  startSearch = textToSearch.find(search);
  while (startSearch != string::npos)
    {
      // Get the iterators.
      matchStart = *startIter;
      gtk_text_iter_forward_chars(&matchStart, startSearch);
      matchEnd = matchStart;
      gtk_text_iter_forward_chars(&matchEnd, searchLength);

      // Move the iterators, if requested.
      switch (searchType)
	{
	case DocCat::WORD:
	  // Move the iterators.
	  if (gtk_text_iter_inside_word(&matchStart))
	    gtk_text_iter_backward_word_start(&matchStart);
	  if (gtk_text_iter_inside_word(&matchEnd))
	    gtk_text_iter_forward_word_end(&matchEnd);
	  break;
	case DocCat::SENTENCE:
	  moveToSentenceBounds(&matchStart, &matchEnd, 0);
	  break;
	case DocCat::SENTENCE2:
	  moveToSentenceBounds(&matchStart, &matchEnd, 2);
	  break;
	case DocCat::STATEMENT:
	  moveToStatementBounds(&matchStart, &matchEnd);
	  break;
	case DocCat::BUFFER:
	  gtk_text_buffer_get_bounds(doc->getTextBuffer(),
				     &matchStart,
				     &matchEnd);
	  break;
	}

      // Code.
      destCat->codeDocText(doc, &matchStart, &matchEnd, Category::UPDATE);

      // Find the next occurance of search.
      startSearch = textToSearch.find(search, startSearch + searchLength);
    }
}

/** Check whether a Category has Coded Text starting on a line in a Document.

    \param doc is the Document

    \param lineNum is the line number

    \return TRUE if the Category has Coded Text starting on the specified
    line, FALSE otherwise.
*/
gboolean
Category::hasLabels(Document *doc, gint lineNum)
{
  DocCodedText *docCodedText;
  GtkTextBuffer *docBuffer;
  GtkTextIter docIter;
  vector<CodedText>::iterator iText;

  g_return_val_if_fail((doc != NULL) && (lineNum >= 0), FALSE);

  // Get the pointer to the DocCodedText structure for this Document.
  docCodedText = &(docCodedTextMap[doc]);

  // Get the Document buffer pointer.
  docBuffer = (docCodedText->doc)->getTextBuffer();

  // Check each Coded Text for this Document.
  for (iText = (docCodedText->codedTextVector).begin();
       iText != (docCodedText->codedTextVector).end();
       iText++)
    {
      // If this Coded Text is marked for deletion, skip it.
      if (iText->ID == Category::PENDING_UNCODE)
	  continue;

      // Get the iterator for the specified line number.
      gtk_text_buffer_get_iter_at_mark(docBuffer,
				       &docIter,
				       iText->docStart);

      // Check if the Coded Text starts on the specified line number.
      if (gtk_text_iter_get_line(&docIter) == lineNum)
	return(TRUE);
    }
  return(FALSE);
}

/** Change the object's properties.

    \param parent is the parent window

    \return true if the properties have changed,
    false otherwise.
 */
bool
Category::changeProperties(GtkWindow *parent)
{
  GtkWidget *table, *hbox;
  GtkWidget *button;
  GtkWidget *fontView;
  GtkTextBuffer *fontBuf;
  GtkTextTag *fontTag;

  PropertyContext fgContext, bgContext, fontContext;
  GtkAttachOptions labelAttach = GTK_SHRINK, textAttach = GTK_FILL;
  GtkWidget *view, *scrolledView;
  GtkTextIter start, end;
  GdkColor fgTempColor, bgTempColor;
  gint result;
  bool status;
  string sProp;
  string prompt;

  // Create the properties window.
  propWindow = gtk_dialog_new_with_buttons("Properties",
					   GTK_WINDOW(mainWindow),
					   GTK_DIALOG_MODAL,
					   GTK_STOCK_CANCEL,
					   GTK_RESPONSE_REJECT,
					   GTK_STOCK_OK,
					   GTK_RESPONSE_ACCEPT,
					   NULL);
  table = gtk_table_new(2, 2, FALSE);

  // Add label and text entry for the name.
  prompt = objName + " name:  ";
  gtk_table_attach(GTK_TABLE(table),
		   gtk_label_new(_(prompt.c_str())),
		   0, 1, 0, 1,
		   labelAttach, labelAttach,
		   5, 5);
  compName = gtk_entry_new();
  gtk_table_attach(GTK_TABLE(table),
		   compName,
		   1, 2, 0, 1,
		   textAttach, textAttach,
		   5, 5);

  // Add label and text buffer for the description.
  prompt = objName + " description:  ";
  gtk_table_attach(GTK_TABLE(table),
		   gtk_label_new(_(prompt.c_str())),
		   0, 1, 1, 2,
		   labelAttach, labelAttach,
		   5, 5);

  // Create view for the description, and put it into a scrolled
  // window.
  view = gtk_text_view_new();
  scrolledView = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledView),
				 GTK_POLICY_AUTOMATIC,
				 GTK_POLICY_AUTOMATIC);
  gtk_container_add(GTK_CONTAINER(scrolledView), view);

  compDesc = gtk_text_view_get_buffer(GTK_TEXT_VIEW(view));
  gtk_table_attach(GTK_TABLE(table),
		   scrolledView,
		   1, 2, 1, 2,
		   textAttach, textAttach,
		   5, 5);

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(propWindow)->vbox),
		     table,
		     FALSE, FALSE, 0);

  // Font view.
  fontView = gtk_text_view_new();
  fontBuf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(fontView));
  gtk_text_buffer_get_bounds(fontBuf, &start, &end);
  gtk_text_buffer_delete(fontBuf, &start, &end);
  gtk_text_buffer_set_text(fontBuf,
			   "abcdefghijk ABCDEFGHIJK",
			   -1);
  fontTag = gtk_text_buffer_create_tag(fontBuf, "tag", NULL);
  gtk_text_buffer_get_bounds(fontBuf, &start, &end);
  gtk_text_buffer_apply_tag(fontBuf,
			    fontTag,
			    &start,
			    &end);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(fontView), FALSE);
  gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(fontView), FALSE);

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(propWindow)->vbox),
		     gtk_hseparator_new(),
		     FALSE, FALSE, 0);


  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(propWindow)->vbox),
		     fontView,
		     FALSE, FALSE, 5);

  // Add buttons.
  hbox = gtk_hbox_new(FALSE, 5);
  // Font.
  button = gtk_button_new_with_mnemonic("Change font");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  fontContext.parent = propWindow;
  fontContext.view = fontView;
  fontContext.tag = fontTag;
  fontContext.font = font;
  g_signal_connect(button, "clicked",
		   G_CALLBACK(changeFont_cb), &fontContext);
  // Foreground.
  button = gtk_button_new_with_mnemonic("Change foreground");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  fgTempColor = fgColor;
  fgContext.parent = propWindow;
  fgContext.view = fontView;
  fgContext.tag = fontTag;
  fgContext.color = &fgTempColor;
  fgContext.property = "foreground-gdk";
  fgContext.font = font;
  g_signal_connect(button, "clicked",
		   G_CALLBACK(changeColor_cb), &fgContext);
  // Background.
  button = gtk_button_new_with_mnemonic("Change background");
  gtk_box_pack_start(GTK_BOX(hbox), button, TRUE, TRUE, 0);
  bgTempColor = bgColor;
  bgContext.parent = propWindow;
  bgContext.view = fontView;
  bgContext.tag = fontTag;
  bgContext.color = &bgTempColor;
  bgContext.property = "background-gdk";
  bgContext.font = font;
  g_signal_connect(button, "clicked",
		   G_CALLBACK(changeColor_cb), &bgContext);

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(propWindow)->vbox),
		     hbox,
		     FALSE, FALSE, 5);

  // Load the properties.
  // Font view.
  g_object_set((gpointer) fontTag,
	       "foreground-gdk", &fgColor,
	       "background-gdk" ,&bgColor,
	       NULL);
  g_object_set((gpointer) fontTag, "font", font.c_str(), NULL);

  sProp = getShortName();
  gtk_entry_set_text(GTK_ENTRY(compName), sProp.c_str());
  gtk_text_buffer_get_bounds(compDesc, &start, &end);
  gtk_text_buffer_delete(compDesc, &start, &end);
  sProp = getDesc();
  gtk_text_buffer_set_text(compDesc, sProp.c_str(), -1);

  // Show the properties window.
  gtk_widget_show_all(propWindow);
  result = gtk_dialog_run(GTK_DIALOG(propWindow));

  // Load new properties.
  if (result == GTK_RESPONSE_ACCEPT)
    {
      // Load the name (pending).
      sProp = gtk_entry_get_text(GTK_ENTRY(compName));
      sProp = getParentsName() + sProp;
      pendingSetName(sProp);

      // Load the font and colors.
      setFont(fontContext.font);
      setFgColor(&fgTempColor);
      setBgColor(&bgTempColor);

      // Load the description.
      gtk_text_buffer_get_bounds(compDesc, &start, &end);
      setDesc(gtk_text_buffer_get_text(compDesc, &start, &end, FALSE));

      status = true;
    }
  else
    status = false;

  gtk_widget_destroy(propWindow);
  propWindow = NULL;
  return(status);
}

/** Create the window to code and view a Category.

    \param parent is the parent window

 */
void
Category::createWindow(GtkWidget *parent)
{
  /* Menus */
  static GnomeUIInfo fileMenu [] = {
    {GNOME_APP_UI_ITEM,
     N_("Export Category"),
     N_("Export the Category to a file"),
     (gpointer) docCatExport_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GNOME_STOCK_PIXMAP_CONVERT,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_SEPARATOR,

    {GNOME_APP_UI_ITEM,
     N_("Print preview"),
     N_("Print preview"),
     (gpointer) printPreview_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GTK_STOCK_PRINT_PREVIEW,
     0, (GdkModifierType) 0, NULL},

    {GNOME_APP_UI_ITEM,
     N_("Print"),
     N_("Print the Document"),
     (gpointer) print_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GTK_STOCK_PRINT,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_SEPARATOR,

    {GNOME_APP_UI_ITEM,
     N_("Close window"),
     N_("Close the Category window"),
     (gpointer) close_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GNOME_STOCK_PIXMAP_CLOSE,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_END
  };

  static GnomeUIInfo catMenu [] = {
    {GNOME_APP_UI_ITEM,
     N_("Select Category"),
     N_("Select the Category for Coding Text"),
     (gpointer) selectCodeCategoryList_cb, NULL, NULL,
     GNOME_APP_PIXMAP_NONE,
     NULL,
     0, (GdkModifierType) 0, NULL},

    {GNOME_APP_UI_ITEM,
     N_("Open Selected Category"),
     N_("Open the Selected Category"),
     (gpointer) codeViewSelectedCat_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GTK_STOCK_OPEN,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_SEPARATOR,

    {GNOME_APP_UI_ITEM,
     N_("Clear Category"),
     N_("Uncode all the Coded Text in this Category"),
     (gpointer) catClearCodedText_cb, NULL, NULL,
     GNOME_APP_PIXMAP_STOCK,
     GTK_STOCK_CLEAR,
     0, (GdkModifierType) 0, NULL},

    GNOMEUIINFO_END
  };

  static GnomeUIInfo help_menu [] = {
    GNOMEUIINFO_HELP ("Category"),
  
    GNOMEUIINFO_END
  };

  static GnomeUIInfo menu [] = {
    { GNOME_APP_UI_SUBTREE_STOCK, N_("File"), NULL, fileMenu, NULL, NULL,
      (GnomeUIPixmapType) 0, NULL, 0,	(GdkModifierType) 0, NULL },

    { GNOME_APP_UI_SUBTREE_STOCK, N_("Category"), NULL, catMenu, NULL, NULL,
      (GnomeUIPixmapType) 0, NULL, 0, (GdkModifierType) 0, NULL },

    GNOMEUIINFO_MENU_HELP_TREE(help_menu),
    GNOMEUIINFO_END
  };

  GtkWidget *status, *vbox, *hbox, *frame;
  GtkWidget *button;
  GtkWidget *scrolledView, *view;
  
  string sProp;

  // Create the main window.
  if (mainWindow == NULL)
    {
      // Create widgets.
      mainWindow = gnome_app_new("Category", _("Category a"));
      gtk_window_set_default_size(GTK_WINDOW(mainWindow), 450, 350);
      gtk_signal_connect(GTK_OBJECT(mainWindow), "enter-notify-event",
			 GTK_SIGNAL_FUNC(categoryUpdate_cb), this);

      status = gnome_appbar_new(FALSE, TRUE, GNOME_PREFERENCES_NEVER);
      frame = gtk_frame_new(NULL);

      gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
      vbox = gtk_vbox_new(FALSE, 5);
      hbox = gtk_hbox_new(FALSE, 5);

      sProp = getName();
      gtk_window_set_title(GTK_WINDOW(mainWindow), sProp.c_str());
      gtk_signal_connect(GTK_OBJECT(mainWindow), "delete-event",
			 GTK_SIGNAL_FUNC(hide_cb), this);

      button = gtk_button_new_with_label("Select");
      gtk_signal_connect(GTK_OBJECT(button), "clicked",
			 GTK_SIGNAL_FUNC(selectCodeCategoryList_cb), this);
      gtk_box_pack_start(GTK_BOX(hbox),
			 button,
			 FALSE, FALSE, 0);

      selectedCatBuffer = gtk_text_buffer_new(projTagTable);
      view = gtk_text_view_new_with_buffer(selectedCatBuffer);
      gtk_text_view_set_editable(GTK_TEXT_VIEW(view),
				 FALSE);
      gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(view),
				       FALSE);

      gtk_box_pack_start(GTK_BOX(hbox),
			 gtk_label_new(_("Selected Category:  ")),
			 FALSE, FALSE, 0);
      gtk_box_pack_start(GTK_BOX(hbox),
			 view,
			 FALSE, FALSE, 0);

      button = gtk_button_new_with_label("Code/Uncode");
      gtk_signal_connect(GTK_OBJECT(button), "clicked",
			 GTK_SIGNAL_FUNC(categoryCode_cb), this);

      gtk_box_pack_start(GTK_BOX(hbox),
			 button,
			 FALSE, FALSE, 0);

      button = gtk_button_new_with_label("Uncode row");
      gtk_signal_connect(GTK_OBJECT(button), "clicked",
			 GTK_SIGNAL_FUNC(categoryUncodeRow_cb), this);

      gtk_box_pack_start(GTK_BOX(hbox),
			 button,
			 FALSE, FALSE, 0);

      gtk_box_pack_start(GTK_BOX(vbox),
			 hbox,
			 FALSE, FALSE, 0);
      gtk_box_pack_start(GTK_BOX(vbox),
			 gtk_hseparator_new(),
			 FALSE, FALSE, 0);

      gtk_text_tag_table_add(projTagTable, catTag);

      codedTextStore = gtk_tree_store_new(Category::NCODED_TEXT_COLUMNS,
					  G_TYPE_STRING,
					  G_TYPE_STRING,
					  G_TYPE_STRING,
					  G_TYPE_STRING,
					  G_TYPE_UINT,
					  G_TYPE_BOOLEAN);

      codedTextView = createCodedTextView(codedTextStore);
      g_signal_connect(G_OBJECT(codedTextView),
		       "button-press-event",
		       G_CALLBACK(categoryButtonPress_cb),
		       this);

      scrolledView = gtk_scrolled_window_new(NULL, NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledView),
				     GTK_POLICY_AUTOMATIC,
				     GTK_POLICY_AUTOMATIC);
      gtk_container_add(GTK_CONTAINER(scrolledView), codedTextView);

      gtk_box_pack_start(GTK_BOX(vbox), scrolledView, TRUE, TRUE, 0);
      loadCodedTextStore(Category::LOAD);

      // Add to application window.
      gnome_app_set_contents(GNOME_APP(mainWindow), frame);
      gnome_app_set_statusbar(GNOME_APP(mainWindow), status);
      gnome_app_create_menus_with_data(GNOME_APP(mainWindow), menu, this);
      gnome_app_install_menu_hints(GNOME_APP(mainWindow), menu);

      // Add the widgets.
      gtk_container_add(GTK_CONTAINER(frame), vbox);
    }
}

/** Code or Uncode selected text in CCT window.

    This method obtains the selected row from the Coded Text view, and
    creates a modal window enabling the user to Code or Uncode the
    Coded Text.
*/
void
Category::codeUncodeCodedText()
{
  GtkWidget *cctWindow; // Code Coded Text window
  GtkWidget *view, *hbox, *vbox, *frame, *button;
  GtkTextBuffer *buffer;
  GtkTextIter textIter, cctIter, cctIter1;
  GtkTreeSelection *selection;
  GtkTreeModel *model;
  GtkTreeIter iter;
  int ID;
  gchar *name;
  Document *doc;
  DocCodedText *docCodedText;
  vector<CodedText>::iterator iText;
  bool found;
  string codedTextString;

  // The window is modal.  If we make the window non-modal, context would
  // need to be part of the Category object.
  static CodedTextContext contextCode, contextUncode;
  gint i;

  g_return_if_fail(codedTextStore != NULL);

  // Check that a Category was selected.
  if (selectedCat == NULL)
    {
      showMessage("You must first select a Category",
		  GTK_MESSAGE_ERROR,
		  mainWindow);
      return;
    }

  // Check that a row of Coded Text was selected.
  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(codedTextView));
  if (!gtk_tree_selection_get_selected(selection, &model, &iter))
    {
      showMessage("You must first select a row to Code",
		  GTK_MESSAGE_ERROR,
		  mainWindow);
      return;
    }

  // Check that a row containing Coded Text was selected.
  // Get the ID.
  gtk_tree_model_get(GTK_TREE_MODEL(codedTextStore),
		     &iter,
		     Category::CODED_TEXT_ID, &ID,
		     -1);

  // Get the Document.
  gtk_tree_model_get(GTK_TREE_MODEL(codedTextStore),
		     &iter,
		     Category::DOCUMENT_NAME, &name,
		     -1);
  doc = (*projectDocMap)[name];
  g_free(name);

  // Get the DocCodedText for this Document.
  docCodedText = &(docCodedTextMap[doc]);
  g_return_if_fail(!docCodedText->cut); // The Document should not be Cut.

  // Search each CodedText for the ID number.
  found = false;
  for (iText = (docCodedText->codedTextVector).begin();
       iText != (docCodedText->codedTextVector).end();
       iText++)
    {
      if (iText->ID == ID)
	{
	  found = true;
	  break;
	}
    }

  // Check that this is a row containing Coded Text.
  if (!found)
    {
      showMessage("You must first select a row with Coded Text to Code",
		  GTK_MESSAGE_ERROR,
		  mainWindow);
      return;
    }

  // Obtain the text.
  gtk_text_buffer_get_iter_at_mark(doc->getTextBuffer(),
				   &(contextCode.docIterStart),
				   iText->docStart);
  gtk_text_buffer_get_iter_at_mark(doc->getTextBuffer(),
				   &(contextCode.docIterEnd),
				   iText->docEnd);
  codedTextString = gtk_text_buffer_get_text(doc->getTextBuffer(),
					     &(contextCode.docIterStart),
					     &(contextCode.docIterEnd),
					     TRUE);

  // Create the window.
  cctWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect(GTK_OBJECT(cctWindow), "destroy",
		     GTK_SIGNAL_FUNC(gtk_widget_destroyed), &cctWindow);
      
  // Set the properties.
  gtk_window_set_modal(GTK_WINDOW(cctWindow), TRUE);
  gtk_window_set_transient_for(GTK_WINDOW(cctWindow),
			       GTK_WINDOW(mainWindow));
  gtk_window_set_title(GTK_WINDOW(cctWindow), "Code Coded Text");

  // The frame goes into the window.
  frame = gtk_frame_new(NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
  gtk_container_add(GTK_CONTAINER(cctWindow), frame);

  vbox = gtk_vbox_new(FALSE, 5);

  // Create the Document widgets.
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox),
		     gtk_label_new(_("Document:  ")),
		     FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox),
		     gtk_label_new((doc->getPrettyName()).c_str()),
		     FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox),
		     hbox,
		     FALSE, FALSE, 0);

  // Create the destination Category widgets.
  hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(hbox),
		     gtk_label_new(_("Selected Category:  ")),
		     FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox),
		     gtk_label_new((selectedCat->getPrettyName()).c_str()),
		     FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox),
		     hbox,
		     FALSE, FALSE, 0);

  gtk_box_pack_start(GTK_BOX(vbox),
		     gtk_hseparator_new(),
		     FALSE, FALSE, 0);

  // Create the text buffer.
  buffer = gtk_text_buffer_new(projTagTable);
  gtk_text_buffer_set_text(buffer, codedTextString.c_str(), -1);
  view = gtk_text_view_new_with_buffer(buffer);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(view), FALSE);
  gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(view), FALSE);
  gtk_box_pack_start(GTK_BOX(vbox),
		     view,
		     TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(vbox),
		     gtk_hseparator_new(),
		     FALSE, FALSE, 0);

  // Apply any destination Category tags to the text buffer.
  textIter = contextCode.docIterStart;
  gtk_text_buffer_get_start_iter(buffer, &cctIter);
  for (i = 0; i < codedTextString.length(); i++)
    {
      if (gtk_text_iter_has_tag(&textIter, selectedCat->getTag()))
	{
	  cctIter1 = cctIter;
	  gtk_text_iter_forward_char(&cctIter1);
	  gtk_text_buffer_apply_tag(buffer,
				    selectedCat->getTag(),
				    &cctIter, &cctIter1);
	}
      gtk_text_iter_forward_char(&textIter);
      gtk_text_iter_forward_char(&cctIter);
    }

  // Load the remainder of the context for Coding.
  contextCode.doc = doc;
  contextCode.cat = selectedCat;
  contextCode.buffer = buffer;
  contextCode.view = view;
  contextCode.code = true;

  // Buttons.
  hbox = gtk_hbox_new(FALSE, 5);
  button = gtk_button_new_with_label("Code");
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
		     GTK_SIGNAL_FUNC(cctCodeUncode_cb), &contextCode);
  gtk_box_pack_start(GTK_BOX(hbox),
		     button,
		     FALSE, FALSE, 0);

  contextUncode = contextCode;
  contextUncode.code = false;
  button = gtk_button_new_with_label("Uncode");
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
		     GTK_SIGNAL_FUNC(cctCodeUncode_cb), &contextUncode);
  gtk_box_pack_start(GTK_BOX(hbox),
		     button,
		     FALSE, FALSE, 0);

  button = gtk_button_new_with_label("Done");
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
		     G_CALLBACK(destroy_cb), cctWindow);
  gtk_box_pack_start(GTK_BOX(hbox),
		     button,
		     FALSE, FALSE, 0);

  gtk_box_pack_start(GTK_BOX(vbox),
		     hbox,
		     FALSE, FALSE, 0);

  // Add to the frame.
  gtk_container_add(GTK_CONTAINER(frame), vbox);

  gtk_widget_show_all(cctWindow);
  gtk_window_present(GTK_WINDOW(cctWindow));
}

/** Uncode the text in a Document.

    \param document is the Document

    \param start is the start of the text to Uncode

    \param end is the end of the text to Uncode

*/
void
Category::uncodeDocText(Document *doc,
			GtkTextIter *start,
			GtkTextIter *end)
{
  GtkTextBuffer *buffer = doc->getTextBuffer();
  GtkTextIter oldStart, oldEnd;

  DocCodedText *docCodedText;
  CodedText codedText;
  vector<CodedText>::iterator iText; // Iter for Coded Text container

  gboolean oldDocument;

  // Has this Document been coded?
  if (docCodedTextMap.count(doc) == 0)
    return;

  // Are the iterators equal?
  if (gtk_text_iter_equal(start, end))
    return;

  docCodedText = &(docCodedTextMap[doc]);
  g_return_if_fail(!docCodedText->cut); // The Document should not be Cut.

  // Check each existing Coded Text.
  for (iText = (docCodedText->codedTextVector).begin();
       iText != (docCodedText->codedTextVector).end();
       iText++)
    {
      // Get limits of existing Coded Text.
      gtk_text_buffer_get_iter_at_mark(buffer,
				       &oldStart,
				       iText->docStart);
      gtk_text_buffer_get_iter_at_mark(buffer,
				       &oldEnd,
				       iText->docEnd);

      // If there's no overlap, do nothing.
      if ((gtk_text_iter_compare(&oldEnd, start) < 0) ||
	  (gtk_text_iter_compare(end , &oldStart) < 0))
	continue;

      // There will be a change.
      setChanged();

      // Start of text to Uncode is within existing Coded Text.
      if ((gtk_text_iter_compare(&oldStart, start) <= 0) &&
	  (gtk_text_iter_compare(start, &oldEnd) <= 0))
      {
	// If end of text to Uncode is also within the existing Coded
	// Text, replace this Coded Text with two Coded Text segments.
	if (gtk_text_iter_compare(end, &oldEnd) <= 0)
	  {
	    // Shorten the existing Coded Text.
	    // Get the new mark.
	    gtk_text_buffer_delete_mark(buffer,
					iText->docEnd);
	    iText->docEnd = gtk_text_buffer_create_mark(buffer,
							NULL,
							start,
							FALSE);
	    // Uncode.
	    gtk_text_buffer_remove_tag(buffer,
				       catTag,
				       start,
				       &oldEnd);

	    // Create new segment.  This adds a new Category label, if
	    // necessary.
	    codeDocText(doc, end, &oldEnd, Category::UPDATE);

	    // We're done checking Coded Text.
	    break;
	  }
	else
	  // The end of text to Uncode is after the end of the
	  // existing Coded Text.  Modify the Coded Text.  No new
	  // Category labels are necessary.
	  {
	    // Get the new mark.
	    gtk_text_buffer_delete_mark(buffer,
					iText->docEnd);
	    iText->docEnd = gtk_text_buffer_create_mark(buffer,
							NULL,
							start,
							FALSE);
	    // Uncode.
	    gtk_text_buffer_remove_tag(buffer,
				       catTag,
				       start,
				       &oldEnd);
	    continue;
	  }
	g_return_if_reached();
      }
      // Start of text to UnCode is before the existing Coded Text.
      if (gtk_text_iter_compare(start, &oldStart) <= 0)
      {
	// End of text to UnCode is after the existing Coded Text.
	if (gtk_text_iter_compare(end, &oldEnd) >= 0)
	  {
	    // The Coded Text is contained entirely within the text to
	    // Uncode.  Flag this Coded Text for deletion.  The
	    // Category labels will be updated when the Coded Text is
	    // Uncoded.
	    iText->ID = Category::PENDING_UNCODE;

	    continue;
	  }
	else
	  // The end of text to Uncode is within the Coded Text.
	  // Modify the Coded Text.
	  {
	    // Get the new mark.
	    gtk_text_buffer_delete_mark(buffer,
					iText->docStart);
	    iText->docStart = gtk_text_buffer_create_mark(buffer,
							  NULL,
							  end,
							  FALSE);
	    // Uncode.
	    gtk_text_buffer_remove_tag(buffer,
				       catTag,
				       &oldStart,
				       end);

	    // Request the Document to update the Category labels.
	    doc->removeCategoryLabel(this, &oldStart);
	    doc->addCategoryLabel(this, end);

	    continue;
	  }
      }
    }

  // Remove rows to be deleted and update the Category view.
  loadCodedTextStore(Category::UPDATE);
}

/** Create the Coded Text view.
 */
GtkWidget *
Category::createCodedTextView(GtkTreeStore *store)
{
  GtkWidget *view;

  // Renderers for the list.
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;

  // Index.
  gint i;

  // Create the view.
  view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));

  /* Create a cell renderer for the first column and arbitrarily make it
   * red for demonstration purposes */
  renderer = gtk_cell_renderer_text_new();
  g_object_set(G_OBJECT(renderer),
	       "foreground", "red",
	       NULL);

  /* Create the first column, associating the "text" attribute of the
   * cell_renderer to the first column of the model */
  column = gtk_tree_view_column_new_with_attributes ("Document/line",
						     renderer,
						     "text", 
						     Category::DOCUMENT_LINE,
						     NULL);
  // Add capabilities.  Click sort.
  gtk_tree_view_column_set_sort_column_id(column, Category::DOCUMENT_LINE);
  gtk_tree_view_column_set_resizable(column, TRUE); // Resizable

  /* Add the column to the view. */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);

  /* Create a cell renderer for the speaker ID, and make it blue. */
  renderer = gtk_cell_renderer_text_new();
  g_object_set(G_OBJECT(renderer),
	       "foreground", "blue",
	       NULL);

  /* Create Speaker ID column. */
  column = gtk_tree_view_column_new_with_attributes ("Speaker ID",
						     renderer,
						     "text", 
						     Category::SPEAKER_ID,
						     NULL);
  // Add capabilities.
  gtk_tree_view_column_set_resizable(column, TRUE); // Resizable

  /* Add the column to the view. */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);

  /* Create a cell renderer for the Coded Text. */
  codedTextRenderer = gtk_cell_renderer_text_new();
  g_object_set_data(G_OBJECT(codedTextRenderer),
		    "column",
		    (gint *) Category::CODED_TEXT);

  /* Create a column, associating the "text" attribute of the
   * cell_renderer to the first column of the model */
  column = gtk_tree_view_column_new_with_attributes ("Coded text",
						     codedTextRenderer,
						     "text",
						     Category::CODED_TEXT,
						     "editable",
						     Category::EDITABLE,
						     NULL);
  // Add capabilities.  Click sort.
  gtk_tree_view_column_set_sort_column_id(column, Category::CODED_TEXT);
  gtk_tree_view_column_set_resizable(column, TRUE); // Resizable

  /* Add the column to the view. */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);

  return(view);
}

/** Load or update the Coded Text in the model.

    \param loadUpdate specifies whether to load or update the model
 */
void
Category::loadCodedTextStore(enum LoadUpdate loadUpdate)
{
  DocCodedText *docCodedText;
  Document *doc;
  vector<CodedText>::iterator iText;

  map<Document *, DocCodedText>::iterator iDoc;
  string docName, docPrettyName;

  // Text buffer variables.
  GtkTextIter startIter, endIter;

  g_return_if_fail(codedTextStore != NULL);

  if (loadUpdate == LOAD)
    // Clear the store.
    gtk_tree_store_clear(codedTextStore);

  // Add Coded Text for each Document.
  for (iDoc = docCodedTextMap.begin(); iDoc != docCodedTextMap.end(); iDoc++)
    {
      docCodedText = &(iDoc->second);

      // Skip this Document if it has been Cut.
      if (docCodedText->cut)
	continue;

      doc = docCodedText->doc;
      docName = doc->getName();
      docPrettyName = doc->getPrettyName();

      // Remove any zero-length Coded Text.
      pruneCodedText(docCodedText);

      // If there is no Coded Text, remove the Document line, erase
      // this docCodedTextMap element, and update the number of
      // Documents column.
      if ((docCodedText->codedTextVector).size() == 0)
	{
	  GtkTreeIter catIter;

	  // Remove the Document line.
	  gtk_tree_store_remove(codedTextStore, &(docCodedText->iter));
	  docCodedTextMap.erase(iDoc);

	  // Update the number of Documents column.
	  componentProjectIter(&catIter);
	  map<const string, Set *>::iterator iSet;

	  // Update the Project store.
	  gtk_tree_store_set(selectCatStore,
			     &catIter,
			     QualResObject::N_DOCS, nDocuments(),
			     -1);

	  // Inform all Sets.
	  for (iSet = projectSetMap->begin();
	       iSet != projectSetMap->end();
	       iSet++)
	    {
	      (iSet->second)->propertiesHaveChanged(this);
	    }
	  continue;
	}

      if (loadUpdate == LOAD)
	// Get a top-level iterator.
	gtk_tree_store_append(codedTextStore,
			      &(docCodedText->iter),
			      NULL);

      // Write Document name.
      gtk_tree_store_set(codedTextStore,
			 &(docCodedText->iter),
			 Category::DOCUMENT_LINE, docPrettyName.c_str(),
			 Category::DOCUMENT_NAME, docName.c_str(),
			 -1);

      // Add each Coded Text.
      for (iText = docCodedText->codedTextVector.begin();
	   iText != docCodedText->codedTextVector.end();
	   iText++)
	{
	  // Get the iterators for the Coded Text.
	  gtk_text_buffer_get_iter_at_mark(doc->getTextBuffer(),
					   &startIter,
					   iText->docStart);
	  gtk_text_buffer_get_iter_at_mark(doc->getTextBuffer(),
					   &endIter,
					   iText->docEnd);

	  if (loadUpdate == LOAD)
	    // Get the iterator for this row.
	    gtk_tree_store_append(codedTextStore,
				  &(iText->catIter),
				  &(docCodedText->iter));
	  
	  // Display the Coded Text in the Document store on the
	  // Category buffer.
	  displayCodedText(doc,
			   &startIter,
			   &endIter,
			   &(iText->catIter),
			   iText->ID);

	}
    }
}

/** Open the Document of the selected Coded Text.

    This method also positions the buffer of the Document so that the
    Coded Text is centered.
*/
void
Category::openSelectedDocument()
{
  GtkTreeModel *model;
  GtkTreeIter iter;
  GtkTreeSelection *selection;
  gchar *name;
  Document *doc;
  DocCodedText *docCodedText;
  vector<CodedText>::iterator iText;
  gint ID;

  g_return_if_fail(codedTextStore != NULL);

  // Get the selected Project Component.
  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(codedTextView));

  // There should be a selection.
  if (!gtk_tree_selection_get_selected(selection, &model, &iter))
    {
      g_return_if_reached();
      return;
    }

  gtk_tree_model_get(GTK_TREE_MODEL(codedTextStore),
		     &iter,
		     Category::DOCUMENT_NAME, &name,
		     Category::CODED_TEXT_ID, &ID,
		     -1);

  // Obtain the Document pointer.
  doc = (*projectDocMap)[name];
  g_free(name);

  // Find the CodedText element.
  docCodedText = &(docCodedTextMap[doc]);
  g_return_if_fail(!docCodedText->cut); // The Document should not be Cut.

  for (iText = docCodedText->codedTextVector.begin();
       iText != docCodedText->codedTextVector.end();
       iText++)
    {
      if (iText->ID == ID)
	{
	  // Make the Document and Coded Text visible.
	  doc->showWindow();
	  doc->makeVisible(iText->docStart);
	  break;
	}
    }
}

/** Update the Coded Text in the model using the Document buffer.

    \param doc is the Document

    \param startTextIter is the iterator for the start of the Coded
    Text in the Document buffer

    \param endTextIter is the iterator for the end of the Coded Text
    in the Document buffer

    \param codedTextIter is the iterator for the row in codedTextStore
    to insert or update

    \param ID is the ID for this row.  This is used to associated a
    CodedText structure with a row.

    This method retrieves the text and line number(s) from textBuf,
    formats it, and adds the information to codedTextBuf.  It also
    applies the tag to the Document buffer textBuf.
*/
void
Category::displayCodedText(Document *doc,
			   GtkTextIter *startTextIter,
			   GtkTextIter *endTextIter,
			   GtkTreeIter *codedTextIter,
			   guint ID)
{
  GtkTextIter statementStart, statementEnd; // Start and end of Statement
  string speakerID; // ID of the Statement (the speaker)
  gchar *text; // Coded text
  string paddedText; // Padded coded text
  gint startNum, endNum, offset; // Line numbers and character offset
  ostrstream lineNumText; // Text line numbers

  gint i;

  g_return_if_fail(codedTextStore != NULL);

  text = gtk_text_buffer_get_text(doc->getTextBuffer(),
				  startTextIter,
				  endTextIter,
				  FALSE);

  // Pad the start with spaces.
  offset = gtk_text_iter_get_line_offset(startTextIter);
  paddedText = text;
  for (i = 0; i < offset; i++)
    paddedText = " " + paddedText;

  // Get the Document line number of the Coded Text.
  startNum = gtk_text_iter_get_line(startTextIter) + 1;
  endNum = gtk_text_iter_get_line(endTextIter) + 1;
  lineNumText << setfill('0');
  lineNumText << setw(LINE_NUMBER_WIDTH) << startNum << "-"
	      << setw(LINE_NUMBER_WIDTH) << endNum << ends;
  lineNumText.rdbuf()->freeze(0);
  lineNumText.seekp(0, ios::beg);

  // Obtain the ID.
  statementStart = *startTextIter;
  statementEnd = *endTextIter;
  moveToStatementBounds(&statementStart, &statementEnd);
  speakerID = getStatementID(&statementStart);
  if (speakerID != "")
    speakerID += ID_DELIMITER_STRING;

  // Load the text and line number.
  gtk_tree_store_set(codedTextStore,
		     codedTextIter,
		     Category::DOCUMENT_LINE, lineNumText.str(),
		     Category::SPEAKER_ID, speakerID.c_str(),
		     Category::CODED_TEXT, paddedText.c_str(),
		     Category::DOCUMENT_NAME, (doc->getName()).c_str(),
		     Category::CODED_TEXT_ID, ID,
		     Category::EDITABLE, TRUE,
		     -1);
  lineNumText.rdbuf()->freeze(0); // Unfreeze memory
  g_free(text);

  // Apply the tag.  This is a workaround for a bug.  When the Coded
  // Text length is 1, then text added to the Coded Text does not
  // appear Coded in the Document buffer.
  gtk_text_buffer_apply_tag(doc->getTextBuffer(),
			    catTag,
			    startTextIter, endTextIter);
}

/***********************************************************************/
/* CALLBACKS */
/***********************************************************************/
/** Code or Uncode text in a Category with CCT window.
*/
void
categoryCode_cb(GtkWidget *widget, gpointer data)
{
  Category *cat = (Category *) data;

  cat->codeUncodeCodedText();
}

/** Uncode a row of text in a Category.
*/
void
categoryUncodeRow_cb(GtkWidget *widget, gpointer data)
{
  Category *cat = (Category *) data;

  cat->uncodeCatSelection();
}

/** Uncode all the Coded Text in a Category.
*/
void
catClearCodedText_cb(GtkWidget *widget, gpointer data)
{
  Category *cat = (Category *) data;
  GtkWidget *dialog;
  gint result;

  // Verify with the user.
  dialog = gtk_message_dialog_new(GTK_WINDOW(cat->getMainWindow()),
				  GTK_DIALOG_DESTROY_WITH_PARENT,
				  GTK_MESSAGE_QUESTION,
				  GTK_BUTTONS_NONE,
				  "Are you sure you want to remove all "
				  "Coded Text from this Category?");
  gtk_dialog_add_buttons(GTK_DIALOG(dialog),
			 GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT,
			 GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
			 NULL);

  result = gtk_dialog_run(GTK_DIALOG(dialog));
  gtk_widget_destroy(dialog);

  if (result == GTK_RESPONSE_ACCEPT)
    cat->clearCodedText();
}

/** Code or Uncode text in a Coded Text window.
*/
void
cctCodeUncode_cb(GtkWidget *widget, gpointer data)
{
  CodedTextContext *context = (CodedTextContext *) data;
  GtkTextIter start, end, cursor;
  gint startOffset, endOffset;

  // Move the focus to the view.
  gtk_widget_grab_focus(context->view);

  // Obtain the Iterators.
  if (!gtk_text_buffer_get_selection_bounds(context->buffer, &start, &end))
    return;

  // Unselect.
  gtk_text_buffer_get_iter_at_mark(context->buffer, 
				   &cursor,
				   gtk_text_buffer_get_insert(context->buffer));
  gtk_text_buffer_place_cursor(context->buffer, &cursor);

  // Apply the tag to the CCT buffer.
  if (context->code == true)
    gtk_text_buffer_apply_tag(context->buffer,
			      (context->cat)->getTag(),
			      &start,
			      &end);
  else
    gtk_text_buffer_remove_tag(context->buffer,
			       (context->cat)->getTag(),
			       &start,
			       &end);

  // Obtain the offsets within the CCT view.
  startOffset = gtk_text_iter_get_offset(&start);
  endOffset = gtk_text_iter_get_offset(&end);

  // Obtain the iterators relative to the Document buffer.
  start = context->docIterStart;
  gtk_text_iter_forward_chars(&start, startOffset);

  end = context->docIterStart;
  gtk_text_iter_forward_chars(&end, endOffset);

  // Code or Uncode.
  if (context->code)
    (context->cat)->codeDocText(context->doc,
				&start,
				&end,
				Category::UPDATE);
  else
    (context->cat)->uncodeDocText(context->doc,
				  &start,
				  &end);
}

/** Update the text model in a Category.
*/
gboolean
categoryUpdate_cb(GtkWidget *widget, GdkEventCrossing *event, gpointer data)
{
  Category *cat = (Category *) data;

  if (event->detail == GDK_NOTIFY_INFERIOR)
    return(true);

  cat->loadCodedTextStore(Category::UPDATE);

  return(true);
}

/** Change the color of a Category.
*/
void
changeColor_cb(GtkWidget *button, gpointer data)
{
  PropertyContext *context = (PropertyContext *) data;
  GtkWidget *dialog;
  GtkColorSelection *colorsel;
  gchar *fontName;
  gint response;
  
  dialog = gtk_color_selection_dialog_new("Changing color");

  gtk_window_set_transient_for(GTK_WINDOW(dialog),
			       GTK_WINDOW(context->parent));
  
  colorsel = GTK_COLOR_SELECTION(GTK_COLOR_SELECTION_DIALOG(dialog)->colorsel);

  gtk_color_selection_set_previous_color(colorsel, context->color);
  gtk_color_selection_set_current_color(colorsel, context->color);
  gtk_color_selection_set_has_palette(colorsel, TRUE);

  response = gtk_dialog_run(GTK_DIALOG(dialog));

  if (response == GTK_RESPONSE_OK)
    {
      gtk_color_selection_get_current_color(colorsel,
					     context->color);
      
      g_object_set((gpointer) context->tag,
		   (context->property).c_str(), context->color,
		   NULL);

      // Try to make the changes visible.  This is a workaround.
      g_object_get((gpointer) context->tag,
		   "font",
		   &fontName,
		   NULL);
      g_object_set((gpointer) context->tag,
		   "font",
		   fontName,
		   NULL);
      g_free(fontName);

    }
  
  gtk_widget_destroy (dialog);
}

/** Change the font of a Category.
*/
void
changeFont_cb(GtkWidget *button, gpointer data)
{
  PropertyContext *context = (PropertyContext *) data;
  GtkWidget *dialog;
  gint response;
  
  dialog = gtk_font_selection_dialog_new("Changing font");

  if (context->font != "")
    gtk_font_selection_dialog_set_font_name(GTK_FONT_SELECTION_DIALOG(dialog),
					    (context->font).c_str());

  gtk_window_set_transient_for(GTK_WINDOW(dialog),
			       GTK_WINDOW(context->parent));
  
  response = gtk_dialog_run(GTK_DIALOG(dialog));

  if (response == GTK_RESPONSE_OK)
    {
      // Load font name.
      context->font =
	gtk_font_selection_dialog_get_font_name(GTK_FONT_SELECTION_DIALOG(dialog));
      // Update font view in properties window.
      g_object_set((gpointer) context->tag,
		   "font",
		   (context->font).c_str(),
		   NULL);
    }
  
  gtk_widget_destroy(dialog);
}

/** Callback to respond to multiple clicks on the Category window.
 */
gboolean
categoryButtonPress_cb(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
  Category *cat = (Category *) data;

  // Double click with button 1.  Open the selected Category.
  if ((event->type == GDK_2BUTTON_PRESS) && (event->button == 1))
    cat->openSelectedDocument();

  // Continue with normal handling.
  return(FALSE);
}
