/*
 * $Id: Category.h,v 1.8 2003/04/01 11:25:33 adeniz Exp $
 * This file is part of qualres
 *
 * Copyright (C) 2002  Baker Research, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */
 
/*
 * See the AUTHORS file for a list of people on the gedit Team.  See the
 * ChangeLog files for a list of changes.
 */

#ifndef CATEGORY_H
#define CATEGORY_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <string>
#include <vector>
#include "qualres.h"
#include "DocCat.h"
#include "Document.h"

/** Structure to hold the Marks for Coded Text.

    This structure has pointers to two GtkTextMarks.  They point to
    the beginning and end of the selected text in the GtkTextBuffer
    of the Document.
*/
struct CodedText
{
  GtkTreeIter catIter; ///< Iterator for codedTextStore (Coded Text row)
  GtkTextMark *docStart; ///< Mark at beginning of Coded Text
  GtkTextMark *docEnd; ///< Mark at end of Coded Text
  guint ID; ///< ID
};

/** Structure to hold all the Coded Text for a Document.

    This structure consists of a container of all the CodedText
    structures for a Document.
*/
struct DocCodedText
{
  GtkTreeIter iter; ///< Iterator for codedTextStore (Document row)
  Document *doc; ///< Document
  vector<CodedText> codedTextVector; ///< All the Coded Text for a Document
  guint nextID; ///< Next available ID for CodedText
  gboolean cut; ///< TRUE if the Document has been Cut
};

/** Class for a Category

    This class implements the methods for a Category

    All access functions of this and derived classes return a pointer
    to const.  For example, getName returns (const string *), pointing
    to the name field.  The object is responsible for this memory; the
    calling program should never free the memory pointed to by this
    pointer.
*/
class Category : public DocCat
{
public:
  /** Coded Text view column numbers.
  */
  enum CodedTextColumns
  {
    DOCUMENT_LINE = 0, // Document pretty name or line number
    SPEAKER_ID, // ID of speaker
    CODED_TEXT, // Padded Coded Text
    DOCUMENT_NAME, // Document full name
    CODED_TEXT_ID, // CodedText ID
    EDITABLE, // Editable flag
    NCODED_TEXT_COLUMNS
  };

  /** Input to method to specify whether to load or update.

      This is used by loadCodedTextStore and codeDocText.
  */
  enum LoadUpdate
  {
    LOAD = 0,
    UPDATE,
    NLOAD_UPDATE
  };

  /** CodedText ID values.
   */
  enum CodedTextIDValues
  {
    PENDING_UNCODE=0,
    INIT_ID,
    NCODED_TEXT_ID_VALUES
  };

  Category() {} ///< Constructor for derived classes.
  Category(GtkTextTagTable *);
  virtual ~Category();

  void setFont(const string &);
  void setFgColor(GdkColor *);
  void setBgColor(GdkColor *);

  virtual void commitName();

  guint nDocuments() const;

  /// Return current font.
  const string *getFont() const { return(&font); }

  /// Return foreground color.
  const GdkColor *getFgColor() const { return(&fgColor); }

  /// Return background color.
  const GdkColor *getBgColor() const { return(&bgColor); }

  GtkTextTag *getTag() const { return(catTag);} ///< Return Category's tag.

 // Printing methods.
  virtual gchar *getTextToPrint() const;
  virtual gchar *getLabeledName() const;

  virtual void codeUncodeText(enum CodeUncode) {} ///< Future expansion.
  virtual void uncodeText() {} ///< Future expansion.
  virtual void codeUncodeCodedText();
  virtual void codeDocText(Document *,
			   GtkTextIter *,
			   GtkTextIter *,
			   Category::LoadUpdate);
  void uncodeDocText(Document *,
		     GtkTextIter *,
		     GtkTextIter *);

  virtual void pendingDelete(QualResObject *);
  virtual void pendingCut(QualResObject *);
  virtual void pendingPaste(QualResObject *);

  void clearCodedText();
  bool codeCatSelection(string &);
  void uncodeCatSelection();
  bool removeRedundant();
  virtual GdomeElement *writeXML(GdomeDocument *)
    throw(FileException);
  virtual bool readXML(GdomeElement *)
    throw(FileException);

  virtual void codeSearchText(string &, DocCat::SearchStore, Category *);
  void codeSearchCodedText(Document *, string &, DocCat::SearchStore,
			   GtkTextIter *, GtkTextIter *, Category *);

  gboolean hasLabels(Document *, gint);

  // GUI methods.
  virtual bool changeProperties(GtkWindow *); ///< Popup properties window.
  virtual void createWindow(GtkWidget *); ///< Create main window.
  void loadCodedTextStore(enum LoadUpdate);
  void openSelectedDocument();
  /// Return a pointer to the buffer containing the Category name.
  GtkTextBuffer *getCatNameBuffer() const { return(catNameBuffer); }

private:
  void init(GtkTextTagTable *);
  void displayCodedText(Document *,
			GtkTextIter *,
			GtkTextIter *,
			GtkTreeIter *,
			guint);

  bool mergeCodedText(DocCodedText *, GtkTextIter *, GtkTextIter *);

  // Text attributes.
  string font; ///< Font
  GdkColor fgColor; ///< Foreground color of Coded Text
  GdkColor bgColor; ///< Background color of Coded Text
  gboolean underlined; ///< True if Coded Text is underlined

protected:
  /// Map of coded text for a Document, using the Document address as the key.
  map<Document *, DocCodedText> docCodedTextMap;

  void markupInit();
  void  pruneCodedText(DocCodedText *);
  GtkWidget *createCodedTextView(GtkTreeStore *store);
  // Widgets.
  GtkWidget *codedTextView; ///< View widget for the Coded Text
  GtkTreeStore *codedTextStore; ///< Model for the Coded Text
  GtkCellRenderer *codedTextRenderer; ///< Renderer for the Coded Text
  GtkTextBuffer *catNameBuffer; ///< Text buffer for holding the Category name

  GtkTextTag *catTag; ///< Category tag
};

// Callbacks.
void categoryCode_cb(GtkWidget *, gpointer);
void categoryUncodeRow_cb(GtkWidget *, gpointer);
void catClearCodedText_cb(GtkWidget *, gpointer);
void cctCodeUncode_cb(GtkWidget *, gpointer);
gboolean categoryUpdate_cb(GtkWidget *, GdkEventCrossing *, gpointer);
gboolean categoryButtonPress_cb(GtkWidget *, GdkEventButton *, gpointer);
void changeColor_cb(GtkWidget *, gpointer);
void changeFont_cb(GtkWidget *, gpointer);

#endif
