/*
 * $Id: Frame.cc,v 1.5 2003/04/02 01:42:46 adeniz Exp $
 * This file is part of qualres
 *
 * Copyright (C) 2002  Baker Research, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */
 
/*
 * See the AUTHORS file for a list of people on the gedit Team.  See the
 * ChangeLog files for a list of changes.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "Frame.h"
#include "NarrativeDocument.h"
#include <strstream>
#include <iomanip>

/** Create empty Frame.
 */
Frame::Frame(GtkTextTagTable *t)
{
  init(t);
}

/** Initialize the object

    \param t is the Project tag table

    This method must be called only once for each object.  A tag
    with the name of the object is created and loaded into the
    Project tag table.
*/
void
Frame::init(GtkTextTagTable *t)
{ 
  objType = QualResObject::FRAME;
  objName = "Frame";
  xmlTag = FRAME_TAG;
  projTagTable = t;
  catTag = gtk_text_tag_new(NULL);
  gtk_text_tag_table_add(projTagTable, catTag);
  catNameBuffer = gtk_text_buffer_new(projTagTable);
  // Create a window that isn't used.
  mainWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  // Create a store for the Framed text.
  codedTextStore = NULL;
#if 0
  codedTextStore = gtk_tree_store_new(NCODED_TEXT_COLUMNS,
				      G_TYPE_STRING,
				      G_TYPE_STRING,
				      G_TYPE_STRING,
				      G_TYPE_STRING,
				      G_TYPE_UINT,
				      G_TYPE_BOOLEAN);
#endif

  // There is no window, so let the methods know that there is no
  // renderer.
  codedTextRenderer = NULL;
  markupInit();
}

/** Initialize the Frame markup properties.
*/
void
Frame::markupInit()
{ 
  GdkColor color;

  color.red = 0;
  color.green = 0;
  color.blue = 0;
  setFgColor(&color);

  color.red = 0xFFFF;
  color.green = 0xFFFF;
  color.blue = 0xFFFF;
  setBgColor(&color);

  setFont(DEFAULT_FONT);
}

/** Destructor.
 */
Frame::~Frame()
{
}

/** Commit the pending new name.

    This function sets the store with additional text.

    pendingSetName must have been called.  If the new name is
    different, the object is marked as changed, the window title is
    changed, and the store containing the Frame name is updated.

*/
void
Frame::commitName()
{
  GtkTextIter start, end;

  Component::commitName();

  // Update the store containing the Category name.
  if (catNameBuffer != NULL)
    {
      gtk_text_buffer_set_text(catNameBuffer,
			       ("Frame:  " + getPrettyName() + "--").c_str(),
			       -1);
      gtk_text_buffer_get_bounds(catNameBuffer,
				 &start,
				 &end);
      gtk_text_buffer_apply_tag(catNameBuffer,
				catTag,
				&start,
				&end);
    }
}

/** Get the Component name labeled with "Frame:  ".

    \returns left header of print output (free with g_free)
**/
gchar *
Frame::getLabeledName() const
{
  return(g_strconcat("Frame:  ", (getPrettyName()).c_str(), NULL));
}

/** Frame text in a Narrative Document

    \param document is the Document

    \param start is the start of the text to code.  start is modified.

    \param end is the end of the text to code.  end is modified.

    \param loadUpdate specifies whether to load or update Coded Text
    container

    This methods first checks that the text to Frame doesn't overlap
    any Framed text.

*/
void
Frame::codeDocText(Document *doc,
		   GtkTextIter *start,
		   GtkTextIter *end,
		   Category::LoadUpdate loadUpdate)
{
  map<const string, Category *>::iterator iFrame;
  Frame *frm;

  GtkTextBuffer *buffer = doc->getTextBuffer();
  GtkTextIter textIter;

  DocCodedText *docCodedText;
  CodedText codedText;
  vector<CodedText>::iterator iter; // Iter for Coded Text container

  gboolean oldDocument;

  GtkWidget *view; /// View for Frame name
  GtkTextChildAnchor *anchor; /// Anchor for the view

  // Return if the iterators are equal.
  if (gtk_text_iter_compare(start, end) == 0)
      return;

  // Check for overlap.
  for (iFrame = projectCatMap->begin();
       iFrame != projectCatMap->end();
       iFrame++)
    {
      frm = (Frame *) iFrame->second;
      if (frm->isFramed(doc, start, end))
	{
	  showMessage("The selected text overlaps or is adjacent to text "
		      "that is already Framed",
		      GTK_MESSAGE_WARNING,
		      doc->getMainWindow());
	}
    }

  // Has this Narrative Document been coded before?
  if (docCodedTextMap.count(doc) == 0)
      oldDocument = FALSE;
  else
      oldDocument = TRUE;

  // Find the DocCodedText structure for this Narrative Document.
  docCodedText = &(docCodedTextMap[doc]);
  if (oldDocument == FALSE)
    {
      docCodedText->nextID = Category::INIT_ID;
      docCodedText->cut = FALSE;

      // Load the Narrative Document pointer.
      docCodedText->doc = doc;
    }

  // Initialize the CodedText for this entry.
  codedText.ID = docCodedText->nextID++;
  codedText.docStart = gtk_text_buffer_create_mark(buffer,
						   NULL,
						   start,
						   TRUE);
  codedText.docEnd = gtk_text_buffer_create_mark(buffer,
						 NULL,
						 end,
						 FALSE);

  // Add to docCodedText for this Document.
  iter = docCodedText->codedTextVector.begin();
  docCodedText->codedTextVector.insert(iter, codedText);

  // Create view for holding Frame name.
  view = gtk_text_view_new_with_buffer(catNameBuffer);

  gtk_text_view_set_editable(GTK_TEXT_VIEW(view), FALSE);
  gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(view), TRUE);

  // Add text and widgets.
  if (loadUpdate == Category::UPDATE)
    gtk_text_buffer_insert(buffer, start, FRAME_TEXT_DELIMITER, -1);
  else
    {
      // Skip over the delimiter.
      gtk_text_iter_forward_chars(start, FRAME_TEXT_DELIMITER_SIZE);
      // Remove the character for the anchor.
      textIter = *start;
      gtk_text_iter_forward_char(&textIter);
      gtk_text_buffer_delete(buffer, start, &textIter);
    }

  // Insert anchor and view.
  anchor = gtk_text_buffer_create_child_anchor(buffer, start);
  gtk_text_view_add_child_at_anchor(GTK_TEXT_VIEW(doc->getTextView()),
				    view,
				    anchor);
  gtk_text_buffer_get_iter_at_mark(buffer, end, codedText.docEnd);

  if (loadUpdate == Category::UPDATE)
    {
      gtk_text_buffer_insert(buffer, end, FRAME_TEXT_DELIMITER, -1);
      gtk_text_buffer_get_iter_at_mark(buffer, start, codedText.docStart);
    }
  else
    gtk_text_iter_backward_chars(end, FRAME_TEXT_DELIMITER_SIZE);

  gtk_widget_show_all(view);

  // Apply the tag.
  gtk_text_buffer_apply_tag(buffer, catTag, start, end);

  // Request the Document to add a Category label.
  //  doc->addCategoryLabel(this, start);

  setChanged();
}

/** Check whether text between a pair of iterators overlaps Framed text.

    \param doc is the Narrative Document containing the text to check

    \param start is the start iterator of the text to check

    \param end is the end iterator of the text to check

    \return TRUE if the the text between start and end is Framed in
    this Frame, or adjacent to Framed text, FALSE otherwise.

*/
gboolean
Frame::isFramed(Document *ndoc, GtkTextIter *start, GtkTextIter *end)
{
  DocCodedText *docCodedText;
  GtkTextBuffer *docBuffer;
  GtkTextIter docIter;
  vector<CodedText>::iterator iText;

  g_return_val_if_fail((ndoc != NULL), FALSE);

  // Return if this Document has not been coded.
  if (docCodedTextMap.count(ndoc) == 0)
    return(FALSE);

  // Get the pointer to the DocCodedText structure for this Narrative
  // Document.
  docCodedText = &(docCodedTextMap[ndoc]);

  // Get the Document buffer pointer.
  docBuffer = (docCodedText->doc)->getTextBuffer();

  // Check each Coded Text for this Document.
  for (iText = (docCodedText->codedTextVector).begin();
       iText != (docCodedText->codedTextVector).end();
       iText++)
    {
      // If this Coded Text is marked for deletion, skip it.
      if (iText->ID == Category::PENDING_UNCODE)
	  continue;

      // If the text to check starts after this Framed text, skip it.
      gtk_text_buffer_get_iter_at_mark(docBuffer,
				       &docIter,
				       iText->docEnd);
      if (gtk_text_iter_compare(start, &docIter) > 0)
	continue;

      // If the text to check ends before this Framed text, skip it.
      gtk_text_buffer_get_iter_at_mark(docBuffer,
				       &docIter,
				       iText->docStart);
      if (gtk_text_iter_compare(end, &docIter) < 0)
	continue;

      // There must be overlap.
      return(TRUE);
    }
  return(FALSE);
}

/** Uncode the text in a Narrative Document.

    \param document is the Narrative Document

    \param textIter points to a position in the Framed text to
    unFrame.
*/
void
Frame::unframeNDocText(Document *doc,
		       GtkTextIter *textIter)
{
  GtkTextBuffer *buffer = doc->getTextBuffer();
  GtkTextIter oldStart, oldEnd, tempIter;

  DocCodedText *docCodedText;
  CodedText codedText;
  vector<CodedText>::iterator iText; // Iter for Coded Text container

  gboolean oldDocument;

  // Has this Document been coded?
  if (docCodedTextMap.count(doc) == 0)
    return;

  docCodedText = &(docCodedTextMap[doc]);
  g_return_if_fail(!docCodedText->cut); // The Document should not be Cut.

  // Check each existing Coded Text.
  for (iText = (docCodedText->codedTextVector).begin();
       iText != (docCodedText->codedTextVector).end();
       iText++)
    {
      // Get limits of existing Coded Text.
      gtk_text_buffer_get_iter_at_mark(buffer,
				       &oldStart,
				       iText->docStart);
      gtk_text_buffer_get_iter_at_mark(buffer,
				       &oldEnd,
				       iText->docEnd);

      // If there's no overlap, do nothing.
      if ((gtk_text_iter_compare(&oldEnd, textIter) < 0) ||
	  (gtk_text_iter_compare(textIter , &oldStart) < 0))
	continue;

      // Set this Framed text to be deleted.
      iText->ID = Category::PENDING_UNCODE;

      // Remove the delimiters and the character for the anchor.
      tempIter = oldStart;
      gtk_text_iter_forward_chars(&tempIter, FRAME_TEXT_DELIMITER_SIZE + 1);
      gtk_text_buffer_delete(buffer, &oldStart, &tempIter);

      gtk_text_buffer_get_iter_at_mark(buffer,
				       &oldEnd,
				       iText->docEnd);
      tempIter = oldEnd;
      gtk_text_iter_backward_chars(&tempIter, FRAME_TEXT_DELIMITER_SIZE);
      gtk_text_buffer_delete(buffer, &oldEnd, &tempIter);


      // Unframe the text.
      pruneCodedText(docCodedText);

      // We're done.
      break;
    }
}
