/*
 * $Id: QualResObject.cc,v 1.5 2003/03/15 14:48:34 adeniz Exp $
 * This file is part of qualres
 *
 * Copyright (C) 2002  Baker Research, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */
 
/*
 * See the AUTHORS file for a list of people on the gedit Team.  See the
 * ChangeLog files for a list of changes.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "QualResObject.h"

/// Constructor.
QualResObject::QualResObject()
{
  mainWindow = NULL;
  // Load the parent window for printing.
  setParentWindow(&mainWindow);
  propWindow = NULL;
  compName = NULL;
  compDesc = NULL;
  nameChanged = false;
  name = "";
  oldName = "";
}

/// Destructor.
QualResObject::~QualResObject()
{
  // Projects do not destroy their main window.
  if (objType == PROJECT)
    return;
  // The parent of propWindow is the Project window, not mainWindow. */
  if (propWindow != NULL)
    gtk_widget_destroy(propWindow);
  if (mainWindow != NULL)
    gtk_widget_destroy(mainWindow);
}

/** Load the object's pending new name.

    \param s contains the new name

    Either commitName or uncommitName must have been called, or this
    must be a newly-created object.  The new name is loaded, but the
    object is not marked as changed until commitName is called.  */
void
QualResObject::pendingSetName(const string &s)
{
  g_return_if_fail(!nameChanged);
  oldName = name;
  nameChanged = true;
  name = s;
}

/** Commit the pending new name.

    pendingSetName must have been called.  If the new name is
    different, the object is marked as changed and the window title is
    changed.  */
void
QualResObject::commitName()
{
  g_return_if_fail(nameChanged);
  if (name != oldName)
    {
      setChanged();
      gtk_window_set_title(GTK_WINDOW(mainWindow), (getPrettyName()).c_str());
    }
  oldName = "";
  nameChanged = false;
}

/** Uncommit the pending new name.

    The original name is restored.  pendingSetName must have been called.
*/
void
QualResObject::uncommitName()
{
  g_return_if_fail(nameChanged);
  name = oldName;
  oldName = "";
  nameChanged = false;
}

/** Return the pretty name of the object.

    \returns the pretty name of the object
*/
string
QualResObject::getPrettyName() const
{
  string prettyName;
  int indx;

  prettyName = name;
  while ((indx = prettyName.find(SEP_STRING)) != string::npos)
    {
      prettyName.erase(indx, SEP_CHAR_LEN);
      prettyName.insert(indx , PRETTY_SEP_STRING);
    }
  return(prettyName);
}

/** Return the short name (the name that's displayed) of a full name.

    \return the short name.  This is the last part of the full name
    separated by the separation character SEP_CHAR.

*/
string
QualResObject::getShortName()
{
  int i;
  i = name.rfind(SEP_CHAR);
  if (i == string::npos)
    return(name);
  else
    return(name.substr(i + 1, name.size()));
}

/** Return the full name less the short name.

    \return the full name less the short name, including the separation
    character

*/
string
QualResObject::getParentsName()
{
  int i;
  i = name.rfind(SEP_CHAR);
  if (i == string::npos)
    return(name.substr(0, 0));
  else
    return(name.substr(0, i + 1));
}

/** Load the object's description property.

    \param d contains the new description
*/
void
QualResObject::setDesc(const string &d)
{
  if (desc != d)
    {
      setChanged();
      desc = d;
    }
}

/** Show the main window.
 */
void
QualResObject::showWindow()
{
  if (mainWindow != NULL)
    {
      gtk_widget_show_all(mainWindow);
      gtk_window_present(GTK_WINDOW(mainWindow));
    }
}

/** Create a Project Component tree view widget

    \param store is the tree model

    \return the pointer to the tree view widget

*/
GtkWidget *
QualResObject::createTreeView(GtkTreeStore *store)
{
  GtkWidget *view;

  // Renderers for the list.
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column, *nameColumn;

  // Index.
  gint i;

  // Create the view.
  view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));

  /* Create a cell renderer for the first column and arbitrarily make it
   * red for demonstration purposes */
  renderer = gtk_cell_renderer_text_new();
  g_object_set(G_OBJECT(renderer),
	       "foreground", "red",
	       NULL);

  /* Create the first column, associating the "text" attribute of the
   * cell_renderer to the first column of the model */
  nameColumn = gtk_tree_view_column_new_with_attributes ("Name", renderer,
							 "text", 0,
							 NULL);
  // Add capabilities.
  gtk_tree_view_column_set_sort_column_id(nameColumn, 0); // Click sort
  gtk_tree_view_column_set_resizable(nameColumn, TRUE); // Resizable

  /* Add the column to the view. */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), nameColumn);

  /* Create a cell render and arbitrarily make it blue for demonstration
   * purposes */
  renderer = gtk_cell_renderer_text_new ();
  g_object_set(G_OBJECT(renderer),
	       "foreground", "blue",
	       NULL);

  /* Create a column, associating the "text" attribute of the
   * cell_renderer to the first column of the model */
  column = gtk_tree_view_column_new_with_attributes ("Description", renderer,
						     "text", 1,
						     NULL);
  // Add capabilities.
  gtk_tree_view_column_set_sort_column_id(column, 1); // Click sort
  gtk_tree_view_column_set_resizable(column, TRUE); // Resizable

  /* Add the column to the view. */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);

  /* Sort in ascending order.  This is a workaround; another column must
     be sorted first. */
  g_signal_emit_by_name((gpointer) column, "clicked");
  g_signal_emit_by_name((gpointer) nameColumn, "clicked");

#if 0
  /* Alternative sort. */
  gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(model),
				       COLUMN_INDEX,
				       GTK_SORT_DESCENDING);
#endif

  return(view);
}

/** Create a Category tree view widget

    \param store is the tree model

    \return the pointer to the tree view widget

*/
GtkWidget *
QualResObject::createCatTreeView(GtkTreeStore *store)
{
  GtkWidget *view;

  // Renderers for the list.
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column, *nameColumn;

  // Index.
  gint i;

  // Create the view.
  view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));

  /* Create a cell renderer for the first column. */
  renderer = gtk_cell_renderer_text_new();

  /* Create the first column, associating the "text" attribute of the
   * cell_renderer to the first column of the model */
  nameColumn =
    gtk_tree_view_column_new_with_attributes ("Name", renderer,
					      "text",
					      QualResObject::SHORT_NAME,
					      "foreground-gdk",
					      QualResObject::FG_COLOR,
					      "background-gdk",
					      QualResObject::BG_COLOR,
					      "font",
					      QualResObject::FONT,
					      NULL);
  // Click sort.
  gtk_tree_view_column_set_sort_column_id(nameColumn,
					  QualResObject::SHORT_NAME);
  gtk_tree_view_column_set_resizable(nameColumn, TRUE); // Resizable

  /* Add the column to the view. */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), nameColumn);

  /* Create a cell render. */
  renderer = gtk_cell_renderer_text_new ();

  /* Create a column, associating the "text" attribute of the
   * cell_renderer to the first column of the model */
  column = gtk_tree_view_column_new_with_attributes("Description", renderer,
						    "text",
						    QualResObject::DESCRIPTION,
						    "foreground-gdk",
						    QualResObject::FG_COLOR,
						    "background-gdk",
						    QualResObject::BG_COLOR,
						    "font",
						    QualResObject::FONT,
						    NULL);
  // Click sort.
  gtk_tree_view_column_set_sort_column_id(column,
					  QualResObject::DESCRIPTION);
  gtk_tree_view_column_set_resizable(column, TRUE); // Resizable

  /* Add the column to the view. */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);

  /* Sort in ascending order.  This is a workaround; another column must
     be sorted first. */
  g_signal_emit_by_name((gpointer) column, "clicked");
  g_signal_emit_by_name((gpointer) nameColumn, "clicked");

  return(view);
}

/** Add a column showing the number of Documents to a Category tree view.

    \param view is the tree view

    \param colNum is the column position (starting from zero)

*/
void
QualResObject::addNDocsColumn(GtkWidget *view, gint colNum)
{
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;

  g_assert(colNum >= 0);

  renderer = gtk_cell_renderer_text_new();
  g_object_set(G_OBJECT(renderer),
	       "xalign", 0.5,
	       NULL);
  column = gtk_tree_view_column_new_with_attributes ("# of Docs", renderer,
						     "text",
						     QualResObject::N_DOCS,
						     "foreground-gdk",
						     QualResObject::FG_COLOR,
						     "background-gdk",
						     QualResObject::BG_COLOR,
						     "font",
						     QualResObject::FONT,
						     NULL);
  gtk_tree_view_column_set_sort_column_id(column,
					  QualResObject::N_DOCS); // Click sort
  gtk_tree_view_column_set_resizable(column, TRUE); // Resizable

  /* Add the column to the view. */
  gtk_tree_view_insert_column(GTK_TREE_VIEW(view), column, colNum);
}

/** Add a column showing a checkbox to Category tree view.

    \param view is the tree view

    \param colNum is the column position (starting from zero)

    \param title is the title of the column
*/
void
QualResObject::addCheckColumn(GtkWidget *view, gint colNum, string title)
{
  GtkCellRenderer *renderer;
  GtkTreeViewColumn *column;

  g_assert(colNum >= 0);

  // Make a checkbox column.
  renderer = gtk_cell_renderer_toggle_new();
  g_object_set(G_OBJECT(renderer),
	       "xalign", 0.0,
	       NULL);
  g_object_set_data(G_OBJECT(renderer),
		    "column",
		    (gint *) QualResObject::CHECK);
  g_signal_connect(G_OBJECT(renderer),
		   "toggled",
		   G_CALLBACK(toggle_cb),
		   gtk_tree_view_get_model(GTK_TREE_VIEW(view)));

  column = gtk_tree_view_column_new_with_attributes (title.c_str(), renderer,
						     "active",
						     QualResObject::CHECK,
						     NULL);

  gtk_tree_view_column_set_clickable (GTK_TREE_VIEW_COLUMN (column), TRUE);

  /* Add the column to the view. */
  gtk_tree_view_insert_column(GTK_TREE_VIEW(view),
			      column,
			      colNum);
}
