/*
 * $Id: SaveObject.cc,v 1.1.1.1 2002/12/31 03:12:45 adeniz Exp $
 * This file is part of qualres
 *
 * Copyright (C) 2002  Baker Research, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */
 
/*
 * See the AUTHORS file for a list of people on the gedit Team.  See the
 * ChangeLog files for a list of changes.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "SaveObject.h"
#include <stdio.h>

/** Create and initialize an XML element in an XML element.

    \param xmlDoc is the XML document that is to contain this element

    \param parentElem is the element that is to contain this element

    \param tag is the tag for the element to be created

    \param text is the text contents of this element


*/
void
SaveObject::setTextInElement(GdomeDocument *xmlDoc,
			     GdomeElement *parentElem,
			     gchar *tag,
			     gchar *text)
  throw(FileException)
{
  GdomeElement *el;
  GdomeText *textNode;
  GdomeException exc;
  GdomeNode *result;
  GdomeDOMString *tempStr, *tempVal;

  tempStr = gdome_str_mkref(tag);
  el = gdome_doc_createElement(xmlDoc, tempStr, &exc);
  gdome_str_unref(tempStr);
  if (el == NULL)
    throw FileException();

  // Create a Text Node.
  tempVal = gdome_str_mkref(text);
  textNode = gdome_doc_createTextNode(xmlDoc, tempVal, &exc);
  gdome_str_unref(tempVal);
  if (textNode == NULL)
    {
      gdome_el_unref(el, &exc);
      throw FileException();
    }

  // Append the Text Node created to the childs list of the
  // text element.
  result = gdome_el_appendChild(el, (GdomeNode *)textNode, &exc);
  if (result != (GdomeNode *) textNode)
    {
      gdome_el_unref(el, &exc);
      gdome_t_unref(textNode, &exc);
      throw FileException();
    }
  gdome_t_unref(textNode, &exc);

  // Append the Text element to the childs list of the parent
  // element.
  result = gdome_el_appendChild(parentElem, (GdomeNode *)el, &exc);
  if (result != (GdomeNode *)el)
    {
      gdome_el_unref(el, &exc);
      throw FileException();
    }
  gdome_el_unref(el, &exc);
}

/** Get the text of a Project Component element.

    \param elem is the element that contains the text element

    \param tag is the tag for the element containing the text element

    \return the contents of the text element of the element with the
    specified tag
*/
string
SaveObject::getTextFromElement(GdomeElement *elem, gchar *tag)
  throw(FileException)
{
  GdomeText *textNode; // The text node
  GdomeElement *textElem; // The text element
  GdomeNodeList *elemList; // List of elements
  gint nElems; // Number of elements
  GdomeException exc;
  GdomeDOMString *tempStr, *tempVal;
  string tempString;
  int i;

  // Get child elements of the Document.
  elemList = gdome_el_childNodes(elem, &exc);
  if (elemList == NULL)
    throw FileException();
  nElems = gdome_nl_length(elemList, &exc);

  for (i = 0; i < nElems; i++)
    {
      // Get element.
      textElem = (GdomeElement *) gdome_nl_item(elemList, i, &exc);
      if (textElem == NULL)
	{
	  gdome_nl_unref(elemList, &exc);
	  throw FileException();
	}

      // If this is not the node with the request tag, go on to the
      // next child.
      tempStr = gdome_el_nodeName(textElem, &exc);
      if (exc)
	{
	  gdome_el_unref(textElem, &exc);
	  gdome_nl_unref(elemList, &exc);
	  throw FileException();
	}
      tempString = tempStr->str;
      gdome_str_unref(tempStr);
      if (tempString != tag)
	{
	  gdome_el_unref(textElem, &exc);
	  continue;
	}

      // Get the text node.
      textNode = (GdomeText *) gdome_el_firstChild(textElem,  &exc);
      if (textNode != NULL)
	{
	  tempVal = gdome_t_data(textNode, &exc);
	  if (exc)
	    {
	      gdome_el_unref(textElem, &exc);
	      gdome_nl_unref(elemList, &exc);
	      gdome_t_unref(textNode, &exc);
	      throw FileException();
	    }

	  tempString = tempVal->str;
	  gdome_str_unref(tempVal);
	  gdome_t_unref(textNode, &exc);
	}
      else
	tempString = "";

      // We found the text.
      gdome_el_unref(textElem, &exc);
      break;
    }

  gdome_nl_unref(elemList, &exc);
  return(tempString);
}

/** Set an attribute in an XML element.

    \param elem is the element containing the attribute

    \param attr is the attribute name

    \param val is the value of the attribute

*/
void
SaveObject::setAttribute(GdomeElement *elem, gchar *attr, gchar *val)
  throw(FileException)
{
  GdomeException exc;
  GdomeDOMString *tempStr, *tempVal;

  tempStr = gdome_str_mkref(attr);
  tempVal = gdome_str_mkref(val);
  gdome_el_setAttribute(elem, tempStr, tempVal, &exc);
  gdome_str_unref(tempStr);
  gdome_str_unref(tempVal);
  if (exc)
    throw FileException();
}

/** Read an attribute from an XML element.


    \param elem is the element containing the attribute

    \param attr is the attribute name

    \return the value of the attrubute
*/
string
SaveObject::getAttribute(GdomeElement *elem, gchar *attr)
  throw(FileException)
{
  GdomeDOMString *tempStr, *tempVal;
  GdomeException exc;
  string tempString;

  tempStr = gdome_str_mkref(attr);
  tempVal = gdome_el_getAttribute (elem, tempStr, &exc);
  gdome_str_unref(tempStr);
  if (exc)
      throw FileException();

  tempString = tempVal->str;
  gdome_str_unref(tempVal);

  return(tempString);
}
