/*
 * $Id: qualresUtility.cc,v 1.9 2003/04/16 14:11:47 adeniz Exp $
 * This file is part of qualres
 *
 * Copyright (C) 2002  Baker Research, Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, 
 * Boston, MA 02111-1307, USA.
 */
 
/*
 * See the AUTHORS file for a list of people on the gedit Team.  See the
 * ChangeLog files for a list of changes.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <vector>
#include <algorithm>
#include "qualresUtility.h"
#include "Component.h"

/** Callback to destroy a widget.
*/
void
destroy_cb(GtkWidget *widget, gpointer data)
{
  gtk_widget_destroy(GTK_WIDGET(data));
}

/** Callback to hide a widget or QualResObject.

    Hides the widget, and returns TRUE so that further processing is
    blocked.
*/
gboolean
hide_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  QualResObject *q = (QualResObject *) data;

  if (q != NULL)
    q->hideWindow();
  else
    gtk_widget_hide(widget);

  return(TRUE);
}

/** Callback to hide a specified widget.

    \param data is the widget to hide

    Hides the widget, and returns TRUE so that further processing is
    blocked.
*/
gboolean
hideData_cb(GtkWidget *widget, gpointer data)
{
  gtk_widget_hide(GTK_WIDGET(data));
  return(TRUE);
}

/** Callback to hide a widget.
*/
void
close_cb(GtkWidget *widget, gpointer data)
{
  QualResObject *q = (QualResObject *) data;

  q->hideWindow();
}

// Dummy callback.
void 
nothing_cb(GtkWidget* widget, gpointer data)
{
  GtkWidget* dialog;
  GtkWidget* app;
  gchar *message;

  message = (gchar *) data;
  if (message == NULL)
    message = "This does nothing; it is only a demonstration";
  //  message = "check";
  g_print("message:  '%s'\n", message);
  app = (GtkWidget*) data;

  dialog = gnome_ok_dialog_parented(message, GTK_WINDOW(app));
}


/** Toggle the checkbox of a view.
 */
void
toggle_cb(GtkCellRendererToggle *cell, gchar *path_str, gpointer data)
{
  GtkTreeModel *model = (GtkTreeModel *)data;
  GtkTreePath *path = gtk_tree_path_new_from_string (path_str);
  GtkTreeIter iter;
  gboolean toggle_item;

  gint *column;

  column = (gint *) g_object_get_data (G_OBJECT (cell), "column");

  /* get toggled iter */
  gtk_tree_model_get_iter (model, &iter, path);
  gtk_tree_model_get (model, &iter, column, &toggle_item, -1);

  /* do something with the value */
  toggle_item ^= 1;

  /* set new value */
  gtk_tree_store_set (GTK_TREE_STORE (model), &iter, column,
		      toggle_item, -1);

  /* clean up */
  gtk_tree_path_free (path);
}

/** Displays a popup dialog.

    \param message is the message to display

    \param type is the type of message (GTK_MESSAGE_INFO, etc)

    \param parent is the parent widget
*/
void
showMessage(string message, GtkMessageType type, GtkWidget *parent)
{
  GtkWidget *dialog;

  dialog = gtk_message_dialog_new(GTK_WINDOW(parent),
				  GTK_DIALOG_DESTROY_WITH_PARENT,
				  type,
				  GTK_BUTTONS_CLOSE,
				  message.c_str());
  gtk_dialog_run(GTK_DIALOG(dialog));
  gtk_widget_destroy(dialog);
  return;
}

/** Move a pair of iterators to the beginning and end of a Statement.

    \param start points to the start of the text

    \param end points to the end of the text

    Upon return, start points to the start of the Statement (the line
    just before start has zero characters, or start is the beginning
    of the buffer), and end points to the end of the Statement (the
    line just after end has zero characters, or end is the end of the
    buffer).
*/
void
moveToStatementBounds(GtkTextIter *start, GtkTextIter *end)
{
  GtkTextIter test;
  // Move to the beginning and end of the lines.
  gtk_text_iter_set_line_offset(start, 0);
  if (!gtk_text_iter_ends_line(end))
    gtk_text_iter_forward_to_line_end(end);

  // Move back one line as long as the previous line has non-zero
  // characters, and is not the beginning of the buffer.
  test = *start;
  gtk_text_iter_backward_line(&test);
  while ((gtk_text_iter_get_chars_in_line(&test) > 1) &&
	 !gtk_text_iter_is_start(start))
    {
      *start = test;
      gtk_text_iter_backward_line(&test);
    }

  // Move forward one line as long as the next line has non-zero
  // characters, and is not the end of the buffer.
  test = *end;
  gtk_text_iter_forward_line(&test);
  if (!gtk_text_iter_ends_line(&test))
    gtk_text_iter_forward_to_line_end(&test);
  while ((gtk_text_iter_get_chars_in_line(&test) > 1) &&
	 !gtk_text_iter_is_end(end))
  {
    *end = test;
    gtk_text_iter_forward_line(&test);
    if (!gtk_text_iter_ends_line(&test))
      gtk_text_iter_forward_to_line_end(&test);
  }
}

/** Move a pair of iterators to the beginning and end of a Sentence.

    \param start points to the start of the text

    \param end points to the end of the text

    \param nContext is the number of Sentences before and after the
    text to include

    Upon return, start points to the start of the Sentence and end
    points to the end of the Sentence.  Sentences can cross line
    boundaries, but not Statement boundaries.

*/
void
moveToSentenceBounds(GtkTextIter *start, GtkTextIter *end, gint nContext)
{
  GtkTextBuffer *docBuffer = gtk_text_iter_get_buffer(start);
  GtkTextBuffer *stmtBuffer; // Buffer for the Statment
  GtkTextIter stmtStart, stmtEnd; // Iterators relative to the Document
  GtkTextIter sntcStart, sntcEnd; // Iterators relative to the Statement
  string stmtText;
  gchar *text;
  gint i;

  // Offsets of search text.
  gint startOffsetInDoc, endOffsetInDoc; // Document relative
  gint startOffsetInStmt, endOffsetInStmt; // Statement relative

  // Obtain the text of the Statement.
  stmtStart = *start;
  stmtEnd = *end;
  moveToStatementBounds(&stmtStart, &stmtEnd);
  text = gtk_text_buffer_get_text(docBuffer,
				  &stmtStart,
				  &stmtEnd,
				  TRUE);
  stmtText = text;
  g_free(text);

  // Replace newlines with space.
  replace(stmtText.begin(), stmtText.end(), NEWLINE_CHAR, ' ');

  // Calculate the offsets.
  startOffsetInDoc = gtk_text_iter_get_offset(start);
  endOffsetInDoc = gtk_text_iter_get_offset(end);
  startOffsetInStmt = startOffsetInDoc - gtk_text_iter_get_offset(&stmtStart);
  endOffsetInStmt = startOffsetInStmt + endOffsetInDoc - startOffsetInDoc;

  // Put the text into a GtkBuffer.
  stmtBuffer = gtk_text_buffer_new(NULL);
  gtk_text_buffer_set_text(stmtBuffer, stmtText.c_str(), -1);

  // Obtain the iterators for the start and end of the search text in the
  // statment buffer.
  gtk_text_buffer_get_start_iter(stmtBuffer, &sntcStart);
  sntcEnd = sntcStart;
  gtk_text_iter_forward_chars(&sntcStart, startOffsetInStmt);
  gtk_text_iter_forward_chars(&sntcEnd, endOffsetInStmt);

  // Move the iterators to the beginning and end of the Sentence.
  if (gtk_text_iter_inside_sentence(&sntcStart))
    gtk_text_iter_backward_sentence_start(&sntcStart);
  if (gtk_text_iter_inside_sentence(&sntcEnd))
    gtk_text_iter_forward_sentence_end(&sntcEnd);

  // Include Context.
  for (i = 0; i < nContext; i++)
    {
      gtk_text_iter_backward_sentence_start(&sntcStart);
      gtk_text_iter_forward_sentence_end(&sntcEnd);
    }

  // Obtain the corresponding offsets.
  startOffsetInStmt = gtk_text_iter_get_offset(&sntcStart);
  endOffsetInStmt = gtk_text_iter_get_offset(&sntcEnd);

  g_object_unref((gpointer) stmtBuffer);

  // Move the iterators.
  *start = stmtStart;
  *end = stmtStart;
  gtk_text_iter_forward_chars(start, startOffsetInStmt);
  gtk_text_iter_forward_chars(end, endOffsetInStmt);
}

/** Get the statement ID.

    \param start points to the start of the Statement

*/
string
getStatementID(GtkTextIter *start)
{
  GtkTextIter end;
  string ID;
  gchar *text;
  gint indx; // Index of ID delimiter

  // Find the end of the line.
  end = *start;
  gtk_text_iter_forward_to_line_end(&end);

  // Get the text.
  text = gtk_text_iter_get_text(start, &end);
  ID = text;
  g_free(text);

  // Search for the ID delimiter.
  indx = ID.find(ID_DELIMITER_STRING);
  if (indx == string::npos)
    ID = "No ID";
  else
    ID = ID.substr(0, indx);

  return(ID);
}

/** Delete a row and all its children  from a GtkTreeStore.

    \param store is the GtkTreeStore

    \param iter points to the row to delete

*/
void
deleteTreeRow(GtkTreeStore *store, GtkTreeIter *iter)
{
  GtkTreeIter childIter;
  gint nChildren, i;

  nChildren = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(store),
					     iter);

  // Delete all the children.
  for (i = nChildren - 1; i >= 0; i--)
    {
      gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(store),
				    &childIter,
				    iter,
				    i);

      deleteTreeRow(store, &childIter);
    }

  // Delete the row.
  gtk_tree_store_remove(store, iter);
}

/** Add the selected Project Component addresses to a container.

    \param model is the tree model with the selections

    \param path is one of the selected path

    \param iter points to one of the selected rows

    \param data is a pointer to the container holding the addresses

    This is a GtkTreeSelectionForeachFunc, called by
    gtk_tree_selection_selected_foreach.
*/
void
addAddressesFromView(GtkTreeModel *model,
		     GtkTreePath *path,
		     GtkTreeIter *iter,
		     gpointer data)
{
  vector<Component *> *selAddresses = (vector<Component *> *) data;
  Component *comp;

  // Get the Component address.
  gtk_tree_model_get(model,
		     iter,
		     QualResObject::ADDRESS,
		     &comp,
		     -1);

  selAddresses->push_back(comp);
}

/** Add the selected Project Component name to a container.

    \param model is the tree model with the selections

    \param path is one of the selected path

    \param iter points to one of the selected rows

    \param data is a pointer to the container holding the names.

    This is a GtkTreeSelectionForeachFunc, called by
    gtk_tree_selection_selected_foreach.
*/
void
addNamesFromView(GtkTreeModel *model,
		 GtkTreePath *path,
		 GtkTreeIter *iter,
		 gpointer data)
{
  map<const string, string> *selectedNames;
  gchar *name;
  
  selectedNames = (map<const string, string> *) data;

  // Get the name;
  gtk_tree_model_get(model,
		     iter,
		     QualResObject::FULL_NAME, &name,
		     -1);
  (*selectedNames)[name] = name;
  g_free(name);
}

/** Add the selected Project Component iterator to a container.

    \param model is the tree model with the selections

    \param path is one of the selected path

    \param iter points to one of the selected rows

    \param data is a pointer to the container holding the names.

    This is a GtkTreeSelectionForeachFunc, called by
    gtk_tree_selection_selected_foreach.
*/
void
addItersFromView(GtkTreeModel *model,
		 GtkTreePath *path,
		 GtkTreeIter *iter,
		 gpointer data)
{
  vector<GtkTreeIter> *selectedIters;
  
  selectedIters = (vector<GtkTreeIter> *) data;

  // Add this iterator.
  selectedIters->push_back(*iter);
}

/** Load a vector of Component addresses.

    This is a GtkTreeForeachFunc, called by gtk_tree_foreach.
*/
gboolean loadComponentAddressesFromStore(GtkTreeModel *model,
					 GtkTreePath *path,
					 GtkTreeIter *iter,
					 gpointer data)
{
  vector<Component *> *addresses = (vector<Component *> *) data;
  Component *comp;

  // Get the Component address.
  gtk_tree_model_get(model,
		     iter,
		     QualResObject::ADDRESS,
		     &comp,
		     -1);

  addresses->push_back(comp);

  // Keep looking.
  return(FALSE);
}

/** Load a vector of checked tree iterators.

    This is a GtkTreeForeachFunc, called by gtk_tree_foreach.
*/
gboolean loadCheckedItersFromStore(GtkTreeModel *model,
				   GtkTreePath *path,
				   GtkTreeIter *iter,
				   gpointer data)
{
  vector<GtkTreeIter> *selectedIters;
  
  selectedIters = (vector<GtkTreeIter> *) data;
  gboolean checked;

  // Get the checked value.
  gtk_tree_model_get(model,
		     iter,
		     QualResObject::CHECK,
		     &checked,
		     -1);

  // Add this iterator.
  if (checked)
    selectedIters->push_back(*iter);

  // Keep looking.
  return(FALSE);
}
